#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Implementation of gas branch."""

from typing import Dict

import click

from ..client.gas import DatasetClientType
from .tbrn import TBRN, TBRNType
from .utility import error, get_dataset_client, get_gas, shorten


def _implement_branch(
    obj: Dict[str, str], tbrn: str, name: str, verbose: bool, is_delete: bool
) -> None:
    info = TBRN(tbrn=tbrn)
    if info.type != TBRNType.DATASET:
        error(f'To operate a branch, "{info}" must be a dataset')

    gas = get_gas(**obj)
    dataset_client = get_dataset_client(gas, info)

    if is_delete:
        _delete_branch(dataset_client, info)
        return

    if name:
        _create_branch(dataset_client, name)
    else:
        _list_branches(dataset_client, verbose)


def _create_branch(dataset_client: DatasetClientType, name: str) -> None:
    if dataset_client.status.is_draft:
        error("Branch cannot be created from a draft")

    if not dataset_client.status.commit_id:
        error(f'To create a branch, "{dataset_client.name}" must have commit history')

    dataset_client.create_branch(name)
    branch_tbrn = TBRN(dataset_client.name, revision=name)
    click.echo(f"{branch_tbrn} has been successfully created")


def _list_branches(dataset_client: DatasetClientType, verbose: bool) -> None:
    branches = dataset_client.list_branches()
    if not verbose:
        for branch in branches:
            click.echo(branch.name)
    else:
        name_length = max(len(branch.name) for branch in branches)
        for branch in branches:
            click.echo(f"{branch.name:{name_length}} {shorten(branch.commit_id)} {branch.title}")


def _delete_branch(dataset_client: DatasetClientType, info: TBRN) -> None:
    if not info.revision:
        error(f'To delete a branch, "{info}" must have a branch name')

    dataset_client._delete_branch(info.revision)  # pylint: disable=protected-access
    click.echo(f"{info} has been successfully deleted")



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Implementation of gas dataset."""

from typing import Dict

import click

from .tbrn import TBRN, TBRNType
from .utility import error, get_gas


def _implement_dataset(obj: Dict[str, str], tbrn: str, is_delete: bool, yes: bool) -> None:
    gas = get_gas(**obj)
    if is_delete:
        if not tbrn:
            error("Missing argument TBRN")

        info = TBRN(tbrn=tbrn)
        if info.type != TBRNType.DATASET:
            error(f'"{tbrn}" is not a dataset')

        if not yes:
            click.confirm(
                f'Dataset "{tbrn}" will be completely deleted.\nDo you want to continue?',
                abort=True,
            )

        gas.delete_dataset(info.dataset_name)
        click.echo(f'Dataset "{tbrn}" is deleted successfully')
        return

    if tbrn:
        info = TBRN(tbrn=tbrn)
        if info.type != TBRNType.DATASET:
            error(f'"{tbrn}" is not a dataset')

        gas.create_dataset(info.dataset_name)
        click.echo(f'Dataset "{tbrn}" is created successfully')
    else:
        for dataset_name in gas.list_dataset_names():
            click.echo(TBRN(dataset_name).get_tbrn())



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Implementation of gas rm."""

from typing import Dict

import click

from .tbrn import TBRN, TBRNType
from .utility import error, get_dataset_client, get_gas


def _implement_rm(obj: Dict[str, str], tbrn: str, is_recursive: bool) -> None:
    gas = get_gas(**obj)
    info = TBRN(tbrn=tbrn)
    dataset_client = get_dataset_client(gas, info, is_fusion=False)

    if info.type not in (TBRNType.SEGMENT, TBRNType.NORMAL_FILE):
        error(f'"{tbrn}" is an invalid path to remove')

    if not info.is_draft:
        error(f'To remove the data, "{info}" must be in draft status, like "{info}#1"')

    if info.type == TBRNType.SEGMENT:
        if not is_recursive:
            error("Please use -r option to remove the whole segment")

        dataset_client.delete_segment(info.segment_name)
    else:
        segment = dataset_client.get_segment(info.segment_name)
        segment.delete_data(info.remote_path)

    click.echo(f"{tbrn} is deleted successfully")



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Implementation of gas commit."""

from typing import Dict, Tuple

import click

from .tbrn import TBRN, TBRNType
from .utility import edit_message, error, format_hint, get_dataset_client, get_gas

_COMMIT_HINT = """
# Please enter the commit message for your changes.
# The Default commit message is set as the draft title.
# Lines starting with '#' will be ignored.
# And an empty commit message aborts the commit.
"""


def _implement_commit(obj: Dict[str, str], tbrn: str, message: Tuple[str, ...]) -> None:
    gas = get_gas(**obj)
    info = TBRN(tbrn=tbrn)
    dataset_client = get_dataset_client(gas, info)

    if info.type != TBRNType.DATASET:
        error(f'To operate a commit, "{info}" must be a dataset')

    if not info.is_draft:
        error(f'To commit, "{info}" must be in draft status, like "{info}#1"')

    dataset_client.checkout(draft_number=info.draft_number)

    draft = dataset_client.get_draft()
    hint_message = format_hint(draft.title, draft.description, _COMMIT_HINT)
    title, description = edit_message(message, hint_message)
    if not title:
        error("Aborting commit due to empty commit message")

    dataset_client.commit(title, description)
    commit_tbrn = TBRN(info.dataset_name, revision=dataset_client.status.commit_id).get_tbrn()
    click.echo(f"Committed successfully: {tbrn} -> {commit_tbrn}")



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Implementation of gas auth."""

from configparser import ConfigParser, NoSectionError
from textwrap import indent
from typing import Dict, Optional
from urllib.parse import urljoin

import click

from ..exception import UnauthorizedError
from .utility import (
    error,
    form_profile_value,
    get_gas,
    is_accesskey,
    read_config,
    read_profile,
    update_config,
    write_config,
)

INDENT = " " * 4


def _implement_auth(  # pylint: disable=too-many-arguments
    obj: Dict[str, str], arg1: str, arg2: str, get: bool, status: bool, unset: bool, is_all: bool
) -> None:
    _check_args_and_options(arg1, arg2, get, status, unset, is_all)
    update_config()
    config_parser = read_config()

    if get:
        _get_auth(obj, config_parser, is_all)
        return

    if status:
        _status_auth(obj, config_parser, is_all)
        return

    if unset:
        _unset_auth(obj, config_parser, is_all)
        return

    if not arg1 and not arg2:
        arg1 = _interactive_auth()

    elif is_accesskey(arg1):
        if _is_gas_url(arg2):
            error('Please use "gas auth [url] [accessKey]" to specify the url and accessKey')
        if arg2:
            error(f'Redundant argument "{arg2}"')

    elif _is_gas_url(arg1):
        if not arg2:
            arg2 = _interactive_auth(arg1)
        elif not is_accesskey(arg2):
            error("Wrong accesskey format")
    else:
        error(f'Invalid argument "{arg1}"')

    _update_profile(config_parser, obj["profile_name"], arg1, arg2)


def _get_auth(obj: Dict[str, str], config_parser: ConfigParser, is_all: bool) -> None:
    if is_all:
        try:
            profiles = config_parser["profiles"]
        except KeyError:
            return
        for key, value in profiles.items():
            _echo_formatted_profile(key, value)
        return

    profile_name = obj["profile_name"]
    try:
        profile = config_parser["profiles"][profile_name]
    except KeyError:
        error(f"Profile '{profile_name}' does not exist.")
    _echo_formatted_profile(profile_name, profile)


def _unset_auth(obj: Dict[str, str], config_parser: ConfigParser, is_all: bool) -> None:
    if is_all:
        config_parser.remove_section("profiles")
    else:
        try:
            removed = config_parser.remove_option("profiles", obj["profile_name"])
        except NoSectionError:
            removed = False
        if not removed:
            error(f"Profile '{obj['profile_name']}' does not exist.")
    write_config(config_parser, show_message=False)
    click.echo("Unset successfully")


def _interactive_auth(url: Optional[str] = None) -> str:
    click.secho(
        "Please visit and login to the TensorBay website to generate your AccessKey", bold=True
    )
    click.secho(
        "Note: TensorBay has multi-regional websites, "
        "please visit the corresponding website based on your location for better experience\n",
        fg="bright_cyan",
    )

    if url:
        developer_url = click.style(urljoin(url, "/tensorbay/developer"), underline=True)
        click.echo(f" > {developer_url}\n")
    else:
        url_cn = click.style("https://gas.graviti.cn/tensorbay/developer", underline=True)
        url_com = click.style("https://gas.graviti.com/tensorbay/developer", underline=True)
        click.echo(f" > {url_com} (Global site)")
        click.echo(f" > {url_cn} (Chinese site)\n")

    access_key = click.prompt(click.style("Paste your AccessKey here", bold=True)).strip()
    if not is_accesskey(access_key):
        error("Wrong accesskey format")
    return access_key  # type: ignore[no-any-return]


def _is_gas_url(arg: str) -> bool:
    return arg.startswith("https://gas.")


def _update_profile(config_parser: ConfigParser, profile_name: str, arg1: str, arg2: str) -> None:
    access_key, url = (arg2, arg1) if arg2 else (arg1, arg2)
    gas_client = get_gas(access_key, url, profile_name)
    try:
        user_info = gas_client.get_user()
    except UnauthorizedError:
        error(f"{access_key} is not a valid AccessKey")

    if not config_parser.has_section("profiles"):
        config_parser.add_section("profiles")
    config_parser["profiles"][profile_name] = form_profile_value(access_key, url)
    write_config(config_parser, show_message=False)

    messages = [
        f"Successfully set authentication info of '{click.style(user_info.name, bold=True)}'"
    ]
    if user_info.team:
        messages.append(f" in '{click.style(user_info.team.name, bold=True)}' team")
    if profile_name != "default":
        messages.append(f" into profile '{click.style(profile_name, bold=True)}'")
    click.echo("".join(messages))


def _check_args_and_options(  # pylint: disable=too-many-arguments
    arg1: str, arg2: str, get: bool, status: bool, unset: bool, is_all: bool
) -> None:
    if is_all and not (get or unset or status):
        error('Use "--all" option with "--get", "--unset" or "--status" option')

    if get + unset + status > 1:
        error('Use at most one of "--get", "--unset" and "--status"')

    if (get or unset or status) and (arg1 or arg2):
        error("Option requires 0 arguments")


def _echo_formatted_profile(name: str, value: str) -> None:
    formatted_value = indent(value, INDENT)
    click.echo(f"{name} = {formatted_value}\n")


def _status_auth(obj: Dict[str, str], config_parser: ConfigParser, is_all: bool) -> None:
    if is_all:
        try:
            profiles = config_parser["profiles"]
        except KeyError:
            return
        for key in profiles:
            _echo_user_info(key, config_parser)
        return

    _echo_user_info(obj["profile_name"], config_parser)


def _echo_user_info(profile_name: str, config_parser: ConfigParser) -> None:
    access_key, url = read_profile(profile_name, config_parser)
    gas_client = get_gas(access_key, url, profile_name)
    try:
        user_info = gas_client.get_user()
    except UnauthorizedError:
        error(f"{access_key} is not a valid AccessKey")

    click.echo(f"{profile_name}\n{INDENT}USER: {user_info.name}")
    team = user_info.team
    if team:
        click.echo(f"{INDENT}TEAM: {team.name}")
    click.echo(f"{INDENT}{access_key}")
    if url:
        click.echo(f"{INDENT}{url}\n")



#!/usr/bin/env python3
#
# Copytright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of the BSTLD dataset."""

from .loader import BSTLD

__all__ = ["BSTLD"]



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Class CloudClient.

The :class:`CloudClient` defines the initial client to interact between local and cloud platform.

"""

from typing import Any, Dict, Iterator, List

from ..dataset import AuthData
from .requests import Client


class CloudClient:
    """:class:`CloudClient` defines the client to interact between local and cloud platform.

    Arguments:
        name: Name of the auth cloud storage config.
        client: The initial client to interact between local and TensorBay.

    """

    def __init__(self, name: str, client: Client) -> None:
        self._name = name
        self._client = client

    def _make_section(self, section: str) -> str:
        return f"cloud/{self._name}/{section}"

    def _get_url(self, file_path: str) -> str:
        params: Dict[str, Any] = {"filePath": file_path}

        response = self._client.open_api_do("GET", self._make_section("files/urls"), params=params)
        return response.json()["url"]  # type: ignore[no-any-return]

    def _list_files(self, path: str, limit: int = 128) -> Iterator[str]:
        params: Dict[str, Any] = {"prefix": path, "limit": limit}

        while True:
            response = self._client.open_api_do(
                "GET", self._make_section("files"), params=params
            ).json()
            yield from response["cloudFiles"]

            if not response["truncated"]:
                break
            params["marker"] = response["nextMarker"]

    def list_auth_data(self, path: str = "") -> List[AuthData]:
        """List all cloud files in the given directory as :class:`~tensorbay.dataset.data.AuthData`.

        Arguments:
            path: The directory path on the cloud platform.

        Returns:
            The list of :class:`~tensorbay.dataset.data.AuthData` of all the cloud files.

        """
        return [
            AuthData(cloud_path, _url_getter=self._get_url) for cloud_path in self._list_files(path)
        ]



#!/usr/bin/env python3
#
# Copyright 2020 Graviti. All Rights Reserved.
#
# pylint: disable=invalid-name

"""This file defines the BDD100K dataloader."""

import json
import os
from glob import glob
from typing import Any, Callable, Dict, Iterable, List
from warnings import warn

from ...dataset import Data, Dataset
from ...label import (
    Classification,
    LabeledBox2D,
    LabeledMultiPolygon,
    LabeledPolygon,
    LabeledPolyline2D,
)
from ...opendataset import _utility

DATASET_NAMES = {
    "100k": "BDD100K",
    "10k": "BDD100K_10K",
    "mots": "BDD100K_MOTS2020",
    "mot": "BDD100K_MOT2020",
}
_SEGMENT_NAMES = ("train", "val")
_LABEL_TYPE_INFO = {
    "100k": {
        "det": ("Detection 2020", "BOX2D"),
        "lane": ("Lane Marking", "POLYLINE2D"),
        "drivable": ("Drivable Area", "POLYGON"),
    },
    "10k": {
        "ins_seg": ("Instance Segmentation", "POLYGON"),
        "sem_seg": ("Semantic Segmentation", "POLYLINE2D"),
        "pan_seg": ("Panoptic Segmentation", "POLYLINE2D"),
    },
}
_TRACKING_DATASET_INFO = {
    "mots": ("bdd100k_seg_track_20", "seg_track_20", os.path.join("seg_track_20", "polygons")),
    "mot": (
        "bdd100k_box_track_20",
        "",
        "",
    ),
}
_DATA_GETTER = Callable[[str, Dict[str, Any]], Data]
_DATA_GENERATOR = Callable[[str, str, _DATA_GETTER], Iterable[Data]]


def BDD100K(path: str) -> Dataset:
    """Load BDD100K Dataset to Tensorbay.

    The file structure should be like::

        <path>
            bdd100k_images_100k/
                images/
                    100k/
                        test
                        train
                        val
                labels/
                    det_20/
                        det_train.json
                        det_val.json
                    lane/
                        polygons/
                            lane_train.json
                            lane_val.json
                    drivable/
                        polygons/
                            drivable_train.json
                            drivable_val.json

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class:`~tensorbay.dataset.dataset.Dataset` instance.

    """
    return _BDD100K_loader(path, "100k")


def _BDD100K_10K(path: str) -> Dataset:
    """Load a sub-dataset 10k of BDD100K Dataset to Tensorbay.

    The dataset is named as 'BDD100K_10K'

    The file structure should be like::

        <path>
            bdd100k_images_10k/
                images/
                    10k/
                        test
                        train
                        val
                labels/
                    ins_seg/
                        polygons/
                            ins_seg_train.json
                            ins_seg_val.json
                    pan_seg/
                        polygons/
                            pan_seg_train.json
                            pan_seg_val.json
                    sem_seg/
                        polygons/
                            sem_seg_train.json
                            sem_seg_val.json

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class:`~tensorbay.dataset.dataset.Dataset` instance.

    """
    return _BDD100K_loader(path, "10k")


def _BDD100K_loader(path: str, dataset_type: str) -> Dataset:
    root_path = os.path.join(
        os.path.abspath(os.path.expanduser(path)), f"bdd100k_images_{dataset_type}"
    )
    dataset = Dataset(DATASET_NAMES[dataset_type])
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), f"catalog_{dataset_type}.json"))

    _load_segment(dataset, root_path, dataset_type)

    return dataset


def _load_segment(dataset: Dataset, root_path: str, dataset_type: str) -> None:
    images_directory = os.path.join(root_path, "images", dataset_type)
    labels_directory = os.path.join(root_path, "labels")

    get_data = _get_data_100k if dataset_type == "100k" else _get_data_10k

    for segment_name in _SEGMENT_NAMES:
        segment = dataset.create_segment(segment_name)
        image_paths = _utility.glob(os.path.join(images_directory, segment_name, "*.jpg"))

        print(f"Reading data to segment '{segment_name}'...")
        if segment_name == "test":
            for image_path in image_paths:
                segment.append(Data(image_path))
        else:
            label_contents = _read_label_file(labels_directory, segment_name, dataset_type)
            for image_path in image_paths:
                segment.append(get_data(image_path, label_contents[os.path.basename(image_path)]))
        print(f"Finished reading data to segment '{segment_name}'")


def _get_data_10k(image_path: str, label_content: Dict[str, Any]) -> Data:
    data = Data(image_path)
    polygon: List[LabeledPolygon] = []
    polyline2d: List[LabeledPolyline2D] = []
    for label_info in label_content["labels"]:
        if "poly2d" in label_info:
            _add_poly2d_label(label_info, polygon, polyline2d)
    data.label.polygon = polygon
    data.label.polyline2d = polyline2d
    return data


def _get_data_100k(image_path: str, label_content: Dict[str, Any]) -> Data:
    data = Data(image_path)
    box2d: List[LabeledBox2D] = []
    polygon: List[LabeledPolygon] = []
    polyline2d: List[LabeledPolyline2D] = []
    data.label.classification = Classification(attributes=label_content["attributes"])
    for label_info in label_content["labels"]:
        if "box2d" in label_info:
            _add_box2d_label(label_info, box2d)
        if "poly2d" in label_info:
            _add_poly2d_label(label_info, polygon, polyline2d)
    data.label.box2d = box2d
    data.label.polygon = polygon
    data.label.polyline2d = polyline2d
    return data


def _add_box2d_label(label_info: Dict[str, Any], box2d: List[LabeledBox2D]) -> None:
    box2d_info = label_info["box2d"]
    labeled_box2d = LabeledBox2D(
        box2d_info["x1"],
        box2d_info["y1"],
        box2d_info["x2"],
        box2d_info["y2"],
        category=label_info["category"],
        attributes=label_info["attributes"],
    )
    box2d.append(labeled_box2d)


def _add_poly2d_label(
    label_info: Dict[str, Any], polygon: List[LabeledPolygon], polyline2d: List[LabeledPolyline2D]
) -> None:
    poly2d_info = label_info["poly2d"][0]
    if poly2d_info["closed"]:
        labeled_polygon = LabeledPolygon(
            points=poly2d_info["vertices"],
            category=label_info["category"],
            attributes=label_info["attributes"],
        )
        polygon.append(labeled_polygon)
    else:
        labeled_polyline2d = LabeledPolyline2D(
            points=poly2d_info["vertices"],
            category=label_info["category"],
            attributes=label_info.get("attributes", {}),
        )
        polyline2d.append(labeled_polyline2d)


def _read_label_file(label_directory: str, segment_name: str, dataset_type: str) -> Dict[str, Any]:
    source_label_contents = []
    label_filenames = glob(
        os.path.join(label_directory, "**", f"*_{segment_name}.json"), recursive=True
    )

    label_type_info = _LABEL_TYPE_INFO[dataset_type]
    label_prefixes = set(label_type_info)
    for label_filename in label_filenames:
        label_file_basename = os.path.basename(label_filename)
        label_prefix = label_file_basename.replace(f"_{segment_name}.json", "")
        try:
            label_prefixes.remove(label_prefix)
        except KeyError:
            warn_message = f"Invalid label file name '{label_file_basename}'! Ignoring.."
            warn(warn_message)
            continue

        label_description = label_type_info[label_prefix][0]
        print(f"Reading '{label_description}' labels to segment '{segment_name}'...")
        with open(label_filename, "r") as fp:
            source_label_contents.append(json.load(fp))
        print(f"Finished reading '{label_description}' labels to segment '{segment_name}'...")

    for missing_label_prefix in label_prefixes:
        warn_message = (
            f"Missing label file '{missing_label_prefix}_{segment_name}.json'! "
            f"The correspondent '{label_type_info[missing_label_prefix][1]}' "
            f"label will be set to empty!"
        )
        warn(warn_message)

    print(f"Merging '{segment_name}' labels...")
    label_contents = _merge_label(source_label_contents)
    print(f"Finished merging '{segment_name}' labels")
    return label_contents


def _merge_label(source_label_contents: List[List[Dict[str, Any]]]) -> Dict[str, Any]:
    label_contents: Dict[str, Any] = {}
    for source_label_content in source_label_contents:
        for image_info in source_label_content:
            image_name = image_info["name"]
            image_label = label_contents.setdefault(image_name, {})
            image_label.setdefault("labels", []).extend(image_info.get("labels", []))
            image_label.setdefault("attributes", {}).update(image_info.get("attributes", {}))

    return label_contents


def _BDD100K_MOTS2020(path: str) -> Dataset:
    """Load a sub-dataset MOTS2020 of BDD100K Dataset to Tensorbay.

    The dataset is named as 'BDD100K_MOTS2020'

    The file structure should be like::

        <path>
            bdd100k_seg_track_20/
                images/
                    seg_track_20/
                        test/
                            cabc30fc-e7726578/
                                cabc30fc-e7726578-0000001.jpg
                                ...
                            ...
                        train/
                            000d4f89-3bcbe37a/
                                000d4f89-3bcbe37a-0000001.jpg
                                ...
                            ...
                        val/
                            b1c9c847-3bda4659/
                                b1c9c847-3bda4659-0000001.jpg
                                ...
                            ...
                labels/
                    seg_track_20/
                        polygons/
                            train/
                                000d4f89-3bcbe37a.json
                                ...
                            val/
                                b1c9c847-3bda4659.json
                                ...

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class:`~tensorbay.dataset.dataset.Dataset` instance.

    """
    return _tracking_loader(path, "mots")


def _BDD100K_MOT2020(path: str) -> Dataset:
    """Load a sub-dataset MOT2020 of BDD100K Dataset to Tensorbay.

    The dataset is named as 'BDD100K_MOT2020'

    The file structure should be like::

        <path>
            bdd100k_box_track_20/
                images/
                    train/
                        00a0f008-3c67908e/
                            00a0f008-3c67908e-0000001.jpg
                            ...
                        ...
                    val/
                        b1c9c847-3bda4659/
                            b1c9c847-3bda4659-0000001.jpg
                            ...
                        ...
                    test/
                        cabc30fc-e7726578/
                            cabc30fc-e7726578-0000001.jpg
                            ...
                        ...
                labels/
                    train/
                        00a0f008-3c67908e.json
                        ...
                    val/
                        b1c9c847-3bda4659.json
                        ...

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class:`~tensorbay.dataset.dataset.Dataset` instance.

    """
    return _tracking_loader(path, "mot")


def _tracking_loader(path: str, tracking_type: str) -> Dataset:
    if tracking_type == "mot":
        get_data = _get_mot_data
    else:
        get_data = _get_mots_data
    root_path = os.path.join(
        os.path.abspath(os.path.expanduser(path)), _TRACKING_DATASET_INFO[tracking_type][0]
    )
    dataset = Dataset(DATASET_NAMES[tracking_type])
    dataset.notes.is_continuous = True
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), f"catalog_{tracking_type}.json"))
    images_directory = os.path.join(root_path, "images", _TRACKING_DATASET_INFO[tracking_type][1])
    labels_directory = os.path.join(root_path, "labels", _TRACKING_DATASET_INFO[tracking_type][2])
    _load_tracking_segment(dataset, images_directory, labels_directory, get_data)

    return dataset


def _load_tracking_segment(
    dataset: Dataset,
    images_directory: str,
    labels_directory: str,
    load_label: _DATA_GETTER,
) -> None:
    for segment_prefix in _SEGMENT_NAMES:
        image_directory = _utility.glob(os.path.join(images_directory, segment_prefix, "*"))
        labels_directory_segment = os.path.join(labels_directory, segment_prefix)
        if segment_prefix == "test":
            generate_data: _DATA_GENERATOR = _generate_test_data
        else:
            generate_data = _generate_data
        for image_subdir in image_directory:
            segment = dataset.create_segment(f"{segment_prefix}_{os.path.basename(image_subdir)}")
            segment.extend(generate_data(image_subdir, labels_directory_segment, load_label))


def _generate_test_data(image_subdir: str, _: str, __: _DATA_GETTER) -> Iterable[Data]:
    yield from map(Data, _utility.glob(os.path.join(image_subdir, "*.jpg")))


def _generate_data(
    image_subdir: str,
    labels_directory_segment: str,
    get_data: _DATA_GETTER,
) -> Iterable[Data]:
    label_filename = f"{os.path.basename(image_subdir)}.json"
    with open(os.path.join(labels_directory_segment, label_filename), "r") as fp:
        label_contents = json.load(fp)
    for label_content in label_contents:
        label_content_name = label_content["name"]
        if "/" in label_content_name:
            label_content_name = label_content_name[len(label_content["videoName"]) + 1 :]
        image_path = os.path.join(image_subdir, label_content_name)

        yield get_data(image_path, label_content)


def _get_mot_data(image_path: str, label_content: Dict[str, Any]) -> Data:
    data = Data(image_path)
    labeled_box2ds = []
    for label_info in label_content.get("labels", ()):
        box2d_info = label_info.get("box2d")
        if not box2d_info:
            continue
        labeled_box2d = LabeledBox2D(
            box2d_info["x1"],
            box2d_info["y1"],
            box2d_info["x2"],
            box2d_info["y2"],
            category=label_info["category"],
            attributes=label_info["attributes"],
            instance=label_info["id"],
        )
        labeled_box2ds.append(labeled_box2d)
    data.label.box2d = labeled_box2ds

    return data


def _get_mots_data(image_path: str, label_content: Dict[str, Any]) -> Data:
    data = Data(image_path)
    labeled_multipolygons = []
    for label_info in label_content.get("labels", ()):
        if "poly2d" not in label_info:
            continue
        labeled_multipolygon = LabeledMultiPolygon(
            polygons=(poly2d_info["vertices"] for poly2d_info in label_info["poly2d"]),
            category=label_info["category"],
            attributes=label_info["attributes"],
            instance=str(label_info["id"]),
        )
        labeled_multipolygons.append(labeled_multipolygon)
    data.label.multi_polygon = labeled_multipolygons

    return data



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Class GAS.

The :class:`GAS` defines the initial client to interact between local and TensorBay.
It provides some operations on datasets level such as :meth:`GAS.create_dataset`,
:meth:`GAS.list_dataset_names` and :meth:`GAS.get_dataset`.

AccessKey is required when operating with dataset.

"""

import logging
from typing import Any, Dict, Generator, Optional, Type, Union, overload

from typing_extensions import Literal

from ..dataset import Dataset, FusionDataset
from ..exception import DatasetTypeError, OperationError, ResourceNotExistError
from .cloud_storage import CloudClient
from .dataset import DatasetClient, FusionDatasetClient
from .lazy import PagingList
from .log import UPLOAD_DATASET_RESUME_TEMPLATE
from .requests import Client, Tqdm
from .status import Status
from .struct import ROOT_COMMIT_ID, UserInfo

DatasetClientType = Union[DatasetClient, FusionDatasetClient]

logger = logging.getLogger(__name__)

DEFAULT_BRANCH = "main"


class GAS:
    """:class:`GAS` defines the initial client to interact between local and TensorBay.

    :class:`GAS` provides some operations on dataset level such as
    :meth:`GAS.create_dataset` :meth:`GAS.list_dataset_names` and :meth:`GAS.get_dataset`.

    Arguments:
        access_key: User's access key.
        url: The host URL of the gas website.

    """

    def __init__(self, access_key: str, url: str = "") -> None:
        self._client = Client(access_key, url)

    def _generate_auth_storage_configs(
        self, name: Optional[str] = None, offset: int = 0, limit: int = 128
    ) -> Generator[Dict[str, Any], None, int]:
        params: Dict[str, Any] = {"offset": offset, "limit": limit}
        if name:
            params["name"] = name

        response = self._client.open_api_do("GET", "auth-storage-configs", "", params=params).json()

        yield from response["configs"]

        return response["totalCount"]  # type: ignore[no-any-return]

    def _get_dataset_with_any_type(self, name: str) -> DatasetClientType:
        info = self._get_dataset(name)
        dataset_id = info["id"]
        is_fusion = info["type"]
        commit_id = info["commitId"]
        default_branch = info["defaultBranch"]
        dataset_alias = info["alias"]

        status = Status(default_branch, commit_id=commit_id)

        ReturnType: Type[DatasetClientType] = FusionDatasetClient if is_fusion else DatasetClient
        return ReturnType(name, dataset_id, self, status=status, alias=dataset_alias)

    def _get_dataset(self, name: str) -> Dict[str, Any]:
        """Get the information of the TensorBay dataset with the input name.

        Arguments:
            name: The name of the requested dataset.

        Returns:
            The dict of dataset information.

        Raises:
            ResourceNotExistError: When the required dataset does not exist.

        """
        if not name:
            raise ResourceNotExistError(resource="dataset", identification=name)

        try:
            response = self._list_datasets(name=name)
            info = response["datasets"][0]
        except IndexError as error:
            raise ResourceNotExistError(resource="dataset", identification=name) from error

        dataset_id = info["id"]
        response = self._client.open_api_do("GET", "", dataset_id).json()
        response["id"] = dataset_id
        return response

    def _list_datasets(
        self,
        name: Optional[str] = None,
        offset: int = 0,
        limit: int = 128,
    ) -> Dict[str, Any]:
        params: Dict[str, Any] = {
            "offset": offset,
            "limit": limit,
        }
        if name:
            params["name"] = name

        response = self._client.open_api_do("GET", "", params=params)
        return response.json()  # type: ignore[no-any-return]

    def _generate_dataset_names(
        self,
        name: Optional[str] = None,
        offset: int = 0,
        limit: int = 128,
    ) -> Generator[str, None, int]:
        response = self._list_datasets(name, offset, limit)

        for item in response["datasets"]:
            yield item["name"]

        return response["totalCount"]  # type: ignore[no-any-return]

    def get_user(self) -> UserInfo:
        """Get the user information with the current accesskey.

        Returns:
            The :class:`.struct.UserInfo` with the current accesskey.

        """
        response = self._client.open_api_do("GET", "users").json()
        return UserInfo.loads(response)

    def get_auth_storage_config(self, name: str) -> Dict[str, Any]:
        """Get the auth storage config with the given name.

        Arguments:
            name: The required auth storage config name.

        Returns:
            The auth storage config with the given name.

        Raises:
            TypeError: When the given auth storage config is illegal.
            ResourceNotExistError: When the required auth storage config does not exist.

        """
        if not name:
            raise TypeError("The given auth storage config name is illegal")

        try:
            config = next(self._generate_auth_storage_configs(name))
        except StopIteration as error:
            raise ResourceNotExistError(
                resource="auth storage config", identification=name
            ) from error

        return config

    def list_auth_storage_configs(self) -> PagingList[Dict[str, Any]]:
        """List auth storage configs.

        Returns:
            The PagingList of all auth storage configs.

        """
        return PagingList(
            lambda offset, limit: self._generate_auth_storage_configs(None, offset, limit),
            128,
        )

    def get_cloud_client(self, name: str) -> CloudClient:
        """Get a cloud client used for interacting with cloud platform.

        Arguments:
            name: The required auth storage config name.

        Returns:
            The cloud client of this dataset.
        """
        return CloudClient(name, self._client)

    @overload
    def create_dataset(
        self,
        name: str,
        is_fusion: Literal[False] = False,
        *,
        region: Optional[str] = None,
        alias: str = "",
    ) -> DatasetClient:
        ...

    @overload
    def create_dataset(
        self,
        name: str,
        is_fusion: Literal[True],
        *,
        region: Optional[str] = None,
        alias: str = "",
    ) -> FusionDatasetClient:
        ...

    @overload
    def create_dataset(
        self,
        name: str,
        is_fusion: bool = False,
        *,
        region: Optional[str] = None,
        alias: str = "",
    ) -> DatasetClientType:
        ...

    def create_dataset(
        self,
        name: str,
        is_fusion: bool = False,
        *,
        region: Optional[str] = None,  # beijing, hangzhou, shanghai
        alias: str = "",
    ) -> DatasetClientType:
        """Create a TensorBay dataset with given name.

        Arguments:
            name: Name of the dataset, unique for a user.
            is_fusion: Whether the dataset is a fusion dataset, True for fusion dataset.
            region: Region of the dataset to be stored,
                only support "beijing", "hangzhou", "shanghai", default is "shanghai".
            alias: Alias of the dataset, default is "".

        Returns:
            The created :class:`~tensorbay.client.dataset.DatasetClient` instance or
            :class:`~tensorbay.client.dataset.FusionDatasetClient` instance (is_fusion=True),
            and the status of dataset client is "commit".

        """
        post_data = {
            "name": name,
            "type": int(is_fusion),  # normal dataset: 0, fusion dataset: 1
            "alias": alias,
        }
        if region:
            post_data["region"] = region

        status = Status(DEFAULT_BRANCH, commit_id=ROOT_COMMIT_ID)

        response = self._client.open_api_do("POST", "", json=post_data)
        ReturnType: Type[DatasetClientType] = FusionDatasetClient if is_fusion else DatasetClient
        return ReturnType(name, response.json()["id"], self, status=status, alias=alias)

    def create_auth_dataset(
        self,
        name: str,
        config_name: str,
        *,
        is_fusion: bool = False,
        alias: str = "",
    ) -> DatasetClientType:
        """Create a TensorBay dataset with given name in auth cloud storage.

        The dataset will be linked to the given auth cloud storage
            and all of relative data will be stored in auth cloud storage.

        Arguments:
            name: Name of the dataset, unique for a user.
            config_name: The auth storage config name.
            is_fusion: Whether the dataset is a fusion dataset, True for fusion dataset.
            alias: Alias of the dataset, default is "".

        Returns:
            The created :class:`~tensorbay.client.dataset.DatasetClient` instance or
            :class:`~tensorbay.client.dataset.FusionDatasetClient` instance (is_fusion=True),
            and the status of dataset client is "commit".

        """
        post_data = {
            "name": name,
            "type": int(is_fusion),  # normal dataset: 0, fusion dataset: 1
            "configName": config_name,
            "alias": alias,
        }

        status = Status(DEFAULT_BRANCH, commit_id=ROOT_COMMIT_ID)

        response = self._client.open_api_do("POST", "", json=post_data)
        ReturnType: Type[DatasetClientType] = FusionDatasetClient if is_fusion else DatasetClient
        return ReturnType(name, response.json()["id"], self, status=status, alias=alias)

    @overload
    def get_dataset(self, name: str, is_fusion: Literal[False] = False) -> DatasetClient:
        ...

    @overload
    def get_dataset(self, name: str, is_fusion: Literal[True]) -> FusionDatasetClient:
        ...

    @overload
    def get_dataset(self, name: str, is_fusion: bool = False) -> DatasetClientType:
        ...

    def get_dataset(self, name: str, is_fusion: bool = False) -> DatasetClientType:
        """Get a TensorBay dataset with given name and commit ID.

        Arguments:
            name: The name of the requested dataset.
            is_fusion: Whether the dataset is a fusion dataset, True for fusion dataset.

        Returns:
            The requested :class:`~tensorbay.client.dataset.DatasetClient` instance or
            :class:`~tensorbay.client.dataset.FusionDatasetClient` instance (is_fusion=True),
            and the status of dataset client is "commit".

        Raises:
            DatasetTypeError: When the requested dataset type is not the same as given.

        """
        info = self._get_dataset(name)

        dataset_id = info["id"]
        type_flag = info["type"]
        commit_id = info["commitId"]
        default_branch = info["defaultBranch"]
        dataset_alias = info["alias"]

        status = Status(default_branch, commit_id=commit_id)

        if is_fusion != type_flag:
            raise DatasetTypeError(name, type_flag)
        ReturnType: Type[DatasetClientType] = FusionDatasetClient if is_fusion else DatasetClient
        return ReturnType(name, dataset_id, self, status=status, alias=dataset_alias)

    def list_dataset_names(self) -> PagingList[str]:
        """List names of all TensorBay datasets.

        Returns:
            The PagingList of all TensorBay dataset names.

        """
        return PagingList(
            lambda offset, limit: self._generate_dataset_names(None, offset, limit), 128
        )

    def update_dataset(self, name: str, *, alias: Optional[str] = None) -> None:
        """Update a TensorBay Dataset.

        Arguments:
            name: Name of the dataset, unique for a user.
            alias: New alias of the dataset.

        """
        dataset_id = self._get_dataset(name)["id"]
        patch_data = {}
        if alias is not None:
            patch_data["alias"] = alias

        if patch_data:
            self._client.open_api_do("PATCH", "", dataset_id, json=patch_data)

    def rename_dataset(self, name: str, new_name: str) -> None:
        """Rename a TensorBay Dataset with given name.

        Arguments:
            name: Name of the dataset, unique for a user.
            new_name: New name of the dataset, unique for a user.

        """
        dataset_id = self._get_dataset(name)["id"]
        patch_data: Dict[str, str] = {"name": new_name}
        self._client.open_api_do("PATCH", "", dataset_id, json=patch_data)

    @overload
    def upload_dataset(
        self,
        dataset: Dataset,
        draft_number: Optional[int] = None,
        *,
        branch_name: Optional[str] = None,
        jobs: int = 1,
        skip_uploaded_files: bool = False,
        quiet: bool = False,
    ) -> DatasetClient:
        ...

    @overload
    def upload_dataset(
        self,
        dataset: FusionDataset,
        draft_number: Optional[int] = None,
        *,
        branch_name: Optional[str] = None,
        jobs: int = 1,
        skip_uploaded_files: bool = False,
        quiet: bool = False,
    ) -> FusionDatasetClient:
        ...

    @overload
    def upload_dataset(
        self,
        dataset: Union[Dataset, FusionDataset],
        draft_number: Optional[int] = None,
        *,
        branch_name: Optional[str] = None,
        jobs: int = 1,
        skip_uploaded_files: bool = False,
        quiet: bool = False,
    ) -> DatasetClientType:
        ...

    def upload_dataset(
        self,
        dataset: Union[Dataset, FusionDataset],
        draft_number: Optional[int] = None,
        *,
        branch_name: Optional[str] = None,
        jobs: int = 1,
        skip_uploaded_files: bool = False,
        quiet: bool = False,
    ) -> DatasetClientType:
        """Upload a local dataset to TensorBay.

        This function will upload all information contains
        in the :class:`~tensorbay.dataset.dataset.Dataset`
        or :class:`~tensorbay.dataset.dataset.FusionDataset`, which includes:

            - Create a TensorBay dataset with the name and type of input local dataset.
            - Upload all :class:`~tensorbay.dataset.segment.Segment`
              or :class:`~tensorbay.dataset.segment.FusionSegment` in the dataset to TensorBay.

        Arguments:
            dataset: The :class:`~tensorbay.dataset.dataset.Dataset` or
                :class:`~tensorbay.dataset.dataset. FusionDataset` needs to be uploaded.
            draft_number: The draft number.
            branch_name: The branch name.
            jobs: The number of the max workers in multi-thread upload.
            skip_uploaded_files: Set it to True to skip the uploaded files.
            quiet: Set to True to stop showing the upload process bar.

        Returns:
            The :class:`~tensorbay.client.dataset.DatasetClient` or
            :class:`~tensorbay.client.dataset.FusionDatasetClient`
            bound with the uploaded dataset.

        Raises:
            OperationError: When uploading the dataset based on both draft number
                and branch name is not allowed.
            Exception: When Exception was raised during uploading dataset.

        """
        dataset_client = self.get_dataset(dataset.name, isinstance(dataset, FusionDataset))

        if draft_number and branch_name:
            raise OperationError(
                "Uploading the dataset based on both draft number and branch name is not allowed"
            )

        if draft_number:
            dataset_client.checkout(draft_number=draft_number)

        else:
            target_branch_name = branch_name if branch_name else dataset_client.status.branch_name

            for draft in dataset_client.list_drafts():
                if draft.branch_name == target_branch_name:
                    dataset_client.checkout(draft_number=draft.number)
                    break
            else:
                dataset_client.create_draft(
                    'Draft autogenerated by "GAS.upload_dataset"', branch_name=target_branch_name
                )

        try:
            if dataset.catalog:
                dataset_client.upload_catalog(dataset.catalog)

            dataset_client.update_notes(**dataset.notes)  # type: ignore[arg-type]

            if isinstance(dataset, Dataset):
                data_count = sum(len(segment) for segment in dataset)
            else:
                data_count = sum(
                    sum(len(frame) for frame in fusion_segment) for fusion_segment in dataset
                )

            with Tqdm(data_count, disable=quiet) as pbar:
                for segment in dataset:
                    dataset_client._upload_segment(  # pylint: disable=protected-access
                        segment,  # type: ignore[arg-type]
                        jobs=jobs,
                        skip_uploaded_files=skip_uploaded_files,
                        pbar=pbar,
                    )
        except Exception:
            if draft_number:
                logger.error(
                    UPLOAD_DATASET_RESUME_TEMPLATE,
                    dataset_client.status.draft_number,
                    dataset_client.status.draft_number,
                )
            raise

        return dataset_client

    def delete_dataset(self, name: str) -> None:
        """Delete a TensorBay dataset with given name.

        Arguments:
            name: Name of the dataset, unique for a user.

        """
        dataset_id = self._get_dataset(name)["id"]
        self._client.open_api_do("DELETE", "", dataset_id)



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""OpenDataset dataloader collections."""

from .AADB import AADB
from .AnimalPose import AnimalPose5, AnimalPose7
from .AnimalsWithAttributes2 import AnimalsWithAttributes2
from .BDD100K import BDD100K
from .BioIDFace import BioIDFace
from .BSTLD import BSTLD
from .CACD import CACD
from .CADC import CADC
from .CarConnection import CarConnection
from .CCPD import CCPD, CCPDGreen
from .CoinImage import CoinImage
from .CompCars import CompCars
from .COVID_CT import COVID_CT
from .COVIDChestXRay import COVIDChestXRay
from .DeepRoute import DeepRoute
from .DogsVsCats import DogsVsCats
from .DownsampledImagenet import DownsampledImagenet
from .Elpv import Elpv
from .FLIC import FLIC
from .Flower import Flower17, Flower102
from .FSDD import FSDD
from .HalpeFullBody import HalpeFullBody
from .HardHatWorkers import HardHatWorkers
from .HeadPoseImage import HeadPoseImage
from .ImageEmotion import ImageEmotionAbstract, ImageEmotionArtphoto
from .JHU_CROWD import JHU_CROWD
from .KenyanFood import KenyanFoodOrNonfood, KenyanFoodType
from .KylbergTexture import KylbergTexture
from .LeedsSportsPose import LeedsSportsPose
from .LISATrafficLight import LISATrafficLight
from .LISATrafficSign import LISATrafficSign
from .NeolixOD import NeolixOD
from .Newsgroups20 import Newsgroups20
from .NightOwls import NightOwls
from .nuScenes import nuScenes
from .RP2K import RP2K
from .THCHS30 import THCHS30
from .THUCNews import THUCNews
from .TLR import TLR
from .UAVDT import UAVDT
from .VOC2012ActionClassification import VOC2012ActionClassification
from .VOC2012Detection import VOC2012Detection
from .WIDER_FACE import WIDER_FACE

__all__ = [
    "nuScenes",
    "AADB",
    "AnimalPose5",
    "AnimalPose7",
    "AnimalsWithAttributes2",
    "BioIDFace",
    "BDD100K",
    "BSTLD",
    "CACD",
    "CADC",
    "CarConnection",
    "CCPD",
    "CCPDGreen",
    "CoinImage",
    "COVIDChestXRay",
    "CompCars",
    "DeepRoute",
    "DogsVsCats",
    "DownsampledImagenet",
    "Elpv",
    "FLIC",
    "FSDD",
    "HalpeFullBody",
    "Flower102",
    "Flower17",
    "HardHatWorkers",
    "HeadPoseImage",
    "ImageEmotionAbstract",
    "ImageEmotionArtphoto",
    "JHU_CROWD",
    "KenyanFoodOrNonfood",
    "KenyanFoodType",
    "KylbergTexture",
    "LISATrafficLight",
    "LISATrafficSign",
    "LeedsSportsPose",
    "NeolixOD",
    "Newsgroups20",
    "NightOwls",
    "RP2K",
    "THCHS30",
    "THUCNews",
    "TLR",
    "UAVDT",
    "WIDER_FACE",
    "COVID_CT",
    "VOC2012Detection",
    "VOC2012ActionClassification",
]



#!/usr/bin/env python3
#
# Copytright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name
# pylint: disable=missing-module-docstring

import os
from collections import defaultdict
from typing import Dict, List
from xml.dom.minidom import parse

from ...dataset import Data, Dataset
from ...label import LabeledBox2D
from .._utility import glob

DATASET_NAME = "TLR"


def TLR(path: str) -> Dataset:
    """Dataloader of the `TLR`_ dataset.

    .. _TLR: http://www.lara.prd.fr/benchmarks/trafficlightsrecognition

    The file structure should like::

        <path>
            root_path/
                Lara3D_URbanSeq1_JPG/
                    frame_011149.jpg
                    frame_011150.jpg
                    frame_<frame_index>.jpg
                    ...
                Lara_UrbanSeq1_GroundTruth_cvml.xml

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class:`~tensorbay.dataset.dataset.Dataset` instance.

    """
    root_path = os.path.abspath(os.path.expanduser(path))

    dataset = Dataset(DATASET_NAME)
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog.json"))
    segment = dataset.create_segment()

    file_paths = glob(os.path.join(root_path, "Lara3D_UrbanSeq1_JPG", "*.jpg"))
    labels = _parse_xml(os.path.join(root_path, "Lara_UrbanSeq1_GroundTruth_cvml.xml"))
    for file_path in file_paths:
        # the image file name looks like:
        # frame_000001.jpg
        frame_index = int(os.path.basename(file_path)[6:-4])
        data = Data(file_path)
        data.label.box2d = labels[frame_index]
        segment.append(data)
    return dataset


def _parse_xml(xml_path: str) -> Dict[int, List[LabeledBox2D]]:
    dom = parse(xml_path)
    frames = dom.documentElement.getElementsByTagName("frame")

    labels = defaultdict(list)

    for frame in frames:
        index = int(frame.getAttribute("number"))
        objects = frame.getElementsByTagName("objectlist")[0]
        for obj in objects.getElementsByTagName("object"):
            box = obj.getElementsByTagName("box")[0]
            box_h = int(box.getAttribute("h"))
            box_w = int(box.getAttribute("w"))
            box_xc = int(box.getAttribute("xc"))
            box_yc = int(box.getAttribute("yc"))

            labels[index].append(
                LabeledBox2D.from_xywh(
                    x=box_xc - box_w // 2,
                    y=box_yc - box_h // 2,
                    width=box_w,
                    height=box_h,
                    category=obj.getElementsByTagName("subtype")[0].firstChild.data,
                )
            )
    return labels



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Class Client and method multithread_upload.

:class:`Client` can send POST, PUT, and GET requests to the TensorBay Dataset Open API.

:meth:`multithread_upload` creates a multi-thread framework for uploading.

"""

import logging
import os
from collections import defaultdict
from concurrent.futures import FIRST_EXCEPTION, ThreadPoolExecutor, wait
from queue import Queue
from threading import Lock
from typing import Any, Callable, DefaultDict, Generic, Iterable, Optional, Tuple, TypeVar
from urllib.parse import urljoin

import urllib3
from requests import Session
from requests.adapters import HTTPAdapter
from requests.exceptions import RequestException
from requests.models import PreparedRequest, Response
from tqdm import tqdm
from urllib3.util.retry import Retry

from ..__version__ import __version__
from ..exception import ResponseError, ResponseErrorDistributor
from .log import RequestLogging, ResponseLogging

logger = logging.getLogger(__name__)


def _get_allowed_methods_keyword() -> str:
    splits = urllib3.__version__.split(".", 2)
    major = int(splits[0])
    minor = int(splits[1])
    return "allowed_methods" if (major, minor) >= (1, 26) else "method_whitelist"


# check the version of urllib3 and choose the correct keyword for "allowed_methods" in "Retry"
_ALLOWED_METHODS = _get_allowed_methods_keyword()


class Config:
    """This is a base class defining the concept of Request Config.

    Attributes:
        max_retries: Maximum retry times of the request.
        allowed_retry_methods: The allowed methods for retrying request.
        allowed_retry_status: The allowed status for retrying request.
            If both methods and status are fitted, the retrying strategy will work.
        timeout: Timeout value of the request in seconds.
        is_internal: Whether the request is from internal.

    """

    def __init__(self) -> None:

        self.max_retries = 3
        self.allowed_retry_methods = ["HEAD", "OPTIONS", "POST", "PUT"]
        self.allowed_retry_status = [429, 500, 502, 503, 504]

        self.timeout = 30
        self.is_internal = False
        self._x_source = "PYTHON-SDK"


config = Config()


class TimeoutHTTPAdapter(HTTPAdapter):
    """This class defines the http adapter for setting the timeout value.

    Arguments:
        *args: Extra arguments to initialize TimeoutHTTPAdapter.
        timeout: Timeout value of the post request in seconds.
        **kwargs: Extra keyword arguments to initialize TimeoutHTTPAdapter.

    """

    def __init__(self, *args: Any, timeout: Optional[int] = None, **kwargs: Any) -> None:
        self.timeout = timeout if timeout is not None else config.timeout
        super().__init__(*args, **kwargs)

    def send(  # pylint: disable=too-many-arguments
        self,
        request: PreparedRequest,
        stream: Any = False,
        timeout: Any = None,
        verify: Any = True,
        cert: Any = None,
        proxies: Any = None,
    ) -> Any:
        """Send the request.

        Arguments:
            request: The PreparedRequest being sent.
            stream: Whether to stream the request content.
            timeout: Timeout value of the post request in seconds.
            verify: A path string to a CA bundle to use or
                a boolean which controls whether to verify the server's TLS certificate.
            cert: User-provided SSL certificate.
            proxies: Proxies dict applying to the request.

        Returns:
            Response object.

        """
        if timeout is None:
            timeout = self.timeout
        return super().send(request, stream, timeout, verify, cert, proxies)


class UserSession(Session):
    """This class defines UserSession."""

    def __init__(self) -> None:
        super().__init__()
        # self.session.hooks["response"] = [logging_hook]

        retry_strategy = Retry(
            total=config.max_retries,
            status_forcelist=config.allowed_retry_status,
            raise_on_status=False,
            **{_ALLOWED_METHODS: config.allowed_retry_methods},
        )

        self.mount("http://", TimeoutHTTPAdapter(20, 20, retry_strategy))
        self.mount("https://", TimeoutHTTPAdapter(20, 20, retry_strategy))

    def request(  # type: ignore[override]  # pylint: disable=signature-differs
        self, method: str, url: str, *args: Any, **kwargs: Any
    ) -> Response:  # noqa: DAR401
        """Make the request.

        Arguments:
            method: Method for the request.
            url: URL for the request.
            *args: Extra arguments to make the request.
            **kwargs: Extra keyword arguments to make the request.

        Returns:
            Response of the request.

        Raises:
            ResponseError: If post response error.

        """
        try:
            response = super().request(method, url, *args, **kwargs)
            if response.status_code not in (200, 201):
                logger.error(
                    "Unexpected status code(%d)!%s", response.status_code, ResponseLogging(response)
                )
                raise ResponseError(response)

            logger.debug(ResponseLogging(response))
            return response

        except RequestException as error:
            logger.error(
                "%s.%s: %s%s",
                error.__class__.__module__,
                error.__class__.__name__,
                error,
                RequestLogging(error.request),
            )
            raise


class Client:
    """This class defines :class:`Client`.

    :class:`Client` defines the client that saves the user and URL information
    and supplies basic call methods that will be used by derived clients,
    such as sending GET, PUT and POST requests to TensorBay Open API.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti gas website.

    """

    _DEFAULT_URL_CN = "https://gas.graviti.cn/"
    _DEFAULT_URL_COM = "https://gas.graviti.com/"

    def __init__(self, access_key: str, url: str = "") -> None:
        if access_key.startswith("Accesskey-"):
            url = url if url else Client._DEFAULT_URL_CN
        elif access_key.startswith("ACCESSKEY-"):
            url = url if url else Client._DEFAULT_URL_COM
        else:
            raise TypeError("Wrong accesskey format!")

        if not url.startswith("https://"):
            raise TypeError("Invalid url, only support url starts with 'https://'")

        self.gateway_url = urljoin(url, "gatewayv2/")
        self.access_key = access_key

        self._sessions: DefaultDict[int, UserSession] = defaultdict(UserSession)
        self._open_api = urljoin(self.gateway_url, "tensorbay-open-api/v1/")

    def _url_make(self, section: str, dataset_id: str = "") -> str:
        """Generate Open API URL.

        Arguments:
            section: The section of the request.
            dataset_id: Dataset ID.

        Returns:
             Open API URL.

        """
        if dataset_id:
            dataset_url = urljoin(self._open_api, "datasets/")
            if section:
                url = urljoin(urljoin(dataset_url, dataset_id + "/"), section)
            else:
                url = urljoin(dataset_url, dataset_id)
        else:
            if section:
                url = urljoin(self._open_api, section)
            else:
                url = urljoin(self._open_api, "datasets")
        return url

    @property
    def session(self) -> UserSession:
        """Create and return a session per PID so each sub-processes will use their own session.

        Returns:
            The session corresponding to the process.
        """
        return self._sessions[os.getpid()]

    def open_api_do(
        self, method: str, section: str, dataset_id: str = "", **kwargs: Any
    ) -> Response:
        """Send a request to the TensorBay Open API.

        Arguments:
            method: The method of the request.
            section: The section of the request.
            dataset_id: Dataset ID.
            **kwargs: Extra keyword arguments to send in the POST request.

        Raises:
            ResponseError: When the status code OpenAPI returns is unexpected.

        Returns:
            Response of the request.

        """
        kwargs.setdefault("headers", {})["X-Token"] = self.access_key
        kwargs["headers"][
            "X-Source"
        ] = f"{config._x_source}/{__version__}"  # pylint: disable=protected-access

        try:
            return self.do(method=method, url=self._url_make(section, dataset_id), **kwargs)
        except ResponseError as error:
            response = error.response
            error_code = response.json()["code"]
            raise ResponseErrorDistributor.get(error_code, ResponseError)(response) from None

    def do(self, method: str, url: str, **kwargs: Any) -> Response:  # pylint: disable=invalid-name
        """Send a request.

        Arguments:
            method: The method of the request.
            url: The URL of the request.
            **kwargs: Extra keyword arguments to send in the GET request.

        Returns:
            Response of the request.

        """
        return self.session.request(method=method, url=url, **kwargs)


_T = TypeVar("_T")


class Tqdm(tqdm):  # type: ignore[misc]
    """A wrapper class of tqdm for showing the process bar.

    Arguments:
        total: The number of excepted iterations.
        disable: Whether to disable the entire progress bar.

    """

    def __init__(self, total: int, disable: bool = False) -> None:
        super().__init__(desc="Uploading", total=total, disable=disable)

    def update_callback(self, _: Any) -> None:
        """Callback function for updating process bar when multithread task is done."""
        self.update()

    def update_for_skip(self, condition: bool) -> bool:
        """Update process bar for the items which are skipped in builtin filter function.

        Arguments:
            condition: The filter condition, the process bar will be updated if condition is False.

        Returns:
            The input condition.

        """
        if not condition:
            self.update()

        return condition


_R = TypeVar("_R")


def multithread_upload(
    function: Callable[[_T], Optional[_R]],
    arguments: Iterable[_T],
    *,
    callback: Optional[Callable[[Tuple[_R, ...]], None]] = None,
    jobs: int = 1,
    pbar: Tqdm,
) -> None:
    """Multi-thread upload framework.

    Arguments:
        function: The upload function.
        arguments: The arguments of the upload function.
        callback: The callback function.
        jobs: The number of the max workers in multi-thread uploading procession.
        pbar: The :class:`Tqdm` instance for showing the upload process bar.

    """
    with ThreadPoolExecutor(jobs) as executor:
        if callback is not None:
            multi_callback = MultiCallbackTask(function=function, callback=callback)
            function = multi_callback.work
        futures = [executor.submit(function, argument) for argument in arguments]

        for future in futures:
            future.add_done_callback(pbar.update_callback)

        done, not_done = wait(futures, return_when=FIRST_EXCEPTION)

        if callback is not None:
            multi_callback.last_callback()

        for future in not_done:
            future.cancel()
        for future in done:
            future.result()


class MultiCallbackTask(Generic[_T, _R]):
    """A class for callbacking in multi-thread work.

    Arguments:
        function: The function of a single thread.
        callback: The callback function.
        size: The size of the task queue to send a callback.

    """

    def __init__(
        self,
        *,
        function: Callable[[_T], Optional[_R]],
        callback: Callable[[Tuple[_R, ...]], None],
        size: int = 50,
    ) -> None:
        self._lock = Lock()
        self._function = function
        self._callback = callback
        self._size = size
        self._arguments: Queue[_R] = Queue()  # pylint: disable=unsubscriptable-object

    def work(self, argument: _T) -> None:
        """Do the work of a single thread.

        Arguments:
            argument: The argument of the function.

        """
        callback_info = self._function(argument)
        if callback_info is not None:
            self._arguments.put(callback_info)
        with self._lock:
            if self._arguments.qsize() >= self._size:
                callback_arguments = tuple(self._arguments.get() for _ in range(self._size))
            else:
                callback_arguments = ()

        if callback_arguments:
            self._callback(callback_arguments)

    def last_callback(self) -> None:
        """Send the last callback when all works have been done."""
        callback_arguments = tuple(self._arguments.get() for _ in range(self._arguments.qsize()))
        if callback_arguments:
            self._callback(callback_arguments)



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Implementation of gas draft."""

from textwrap import indent
from typing import Dict, Optional, Tuple

import click

from ..client.gas import DatasetClientType
from ..client.struct import ROOT_COMMIT_ID
from ..exception import ResourceNotExistError
from .auth import INDENT
from .tbrn import TBRN, TBRNType
from .utility import edit_message, error, format_hint, get_dataset_client, get_gas

_DRAFT_HINT = """
# Please enter the message for your draft.
# Lines starting with '#' will be ignored.
# And an empty draft message aborts the creation.
"""

_FULL_DRAFT_MESSAGE = """Branch:{}{}

    {}
"""


def _implement_draft(  # pylint: disable=too-many-arguments
    obj: Dict[str, str], tbrn: str, is_list: bool, edit: bool, close: bool, message: Tuple[str, ...]
) -> None:
    gas = get_gas(**obj)
    info = TBRN(tbrn=tbrn)
    dataset_client = get_dataset_client(gas, info)

    if info.type != TBRNType.DATASET:
        error(f'To operate a draft, "{info}" must be a dataset')

    if is_list:
        _list_drafts(dataset_client, info)
    elif edit:
        _edit_draft(dataset_client, info, message)
    elif close:
        _close_draft(dataset_client, info)
    else:
        _create_draft(dataset_client, info, message)


def _create_draft(dataset_client: DatasetClientType, info: TBRN, message: Tuple[str, ...]) -> None:
    if info.is_draft:
        error(f'Create a draft in draft status "{info}" is not permitted')

    title, description = edit_message(message, _DRAFT_HINT)
    if not title:
        error("Aborting creating draft due to empty draft message")

    dataset_client.create_draft(title=title, description=description)
    status = dataset_client.status
    draft_tbrn = TBRN(info.dataset_name, draft_number=status.draft_number).get_tbrn()
    click.echo(f"{draft_tbrn} is created successfully")
    _echo_draft(dataset_client, title, description, status.branch_name)


def _list_drafts(dataset_client: DatasetClientType, info: TBRN) -> None:
    if info.revision:
        error(f'list drafts based on given revision "{info}" is not supported')

    if info.is_draft:
        draft = dataset_client.get_draft(info.draft_number)
        click.echo(f"Draft: {info.get_tbrn()}")
        _echo_draft(dataset_client, draft.title, draft.description, draft.branch_name)
    else:
        for draft in dataset_client.list_drafts():
            click.echo(f"Draft: {TBRN(info.dataset_name, draft_number=draft.number).get_tbrn()}")
            _echo_draft(dataset_client, draft.title, draft.description, draft.branch_name)


def _echo_draft(
    dataset_client: DatasetClientType,
    title: str = "",
    description: str = "",
    branch_name: Optional[str] = None,
) -> None:
    if not branch_name:
        error("Draft should be created based on a branch.")

    try:
        branch = dataset_client.get_branch(branch_name)
    except ResourceNotExistError:
        error(f'The branch "{branch_name}" does not exist')

    if branch.commit_id != ROOT_COMMIT_ID:
        commit_id = f"({branch.commit_id})"
    else:
        commit_id = ""

    if not title:
        title = "<no title>"
    if description:
        description = f"\n\n{indent(description, INDENT)}"
    draft_message = f"{title}{description}"
    click.echo(
        _FULL_DRAFT_MESSAGE.format(
            branch_name,
            commit_id,
            draft_message,
        )
    )


def _edit_draft(dataset_client: DatasetClientType, info: TBRN, message: Tuple[str, ...]) -> None:
    if not info.is_draft:
        error("Draft number is required when editing draft")

    draft = dataset_client.get_draft()
    hint_message = format_hint(draft.title, draft.description, _DRAFT_HINT)
    title, description = edit_message(message, hint_message)
    if not title:
        error("Aborting updating draft due to empty draft message")

    dataset_client.update_draft(title=title, description=description)
    click.echo(f"{info.get_tbrn()} is updated successfully!")
    _echo_draft(dataset_client, title, description, dataset_client.status.branch_name)


def _close_draft(dataset_client: DatasetClientType, info: TBRN) -> None:
    if not info.is_draft:
        error("Draft number is required when editing draft")

    dataset_client.close_draft()
    click.echo(f"{info.get_tbrn()} is closed.")



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of the THUCNews dataset."""

from .loader import THUCNews

__all__ = ["THUCNews"]



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of the THCHS-30 dataset."""

from .loader import THCHS30

__all__ = ["THCHS30"]



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name
# pylint: disable=missing-module-docstring

import json
import os
from typing import Iterable, Iterator, List, Tuple

from ...dataset import Data, Dataset
from ...exception import ModuleImportError
from ...geometry import Keypoint2D
from ...label import LabeledBox2D, LabeledKeypoints2D
from .._utility import glob

DATASET_NAME_5 = "AnimalPose5"
DATASET_NAME_7 = "AnimalPose7"

_KEYPOINT_TO_INDEX = {
    "L_Eye": 0,
    "R_Eye": 1,
    "Nose": 4,
    "L_EarBase": 2,
    "R_EarBase": 3,
    "Throat": 5,
    "L_F_Elbow": 8,
    "R_F_Elbow": 9,
    "L_F_Paw": 16,
    "R_F_Paw": 17,
    "Withers": 7,
    "TailBase": 6,
    "L_B_Paw": 18,
    "R_B_Paw": 19,
    "L_B_Elbow": 10,
    "R_B_Elbow": 11,
    "L_F_Knee": 12,
    "R_F_Knee": 13,
    "L_B_Knee": 14,
    "R_B_Knee": 15,
}


def AnimalPose5(path: str) -> Dataset:
    """Dataloader of the `5 Categories Animal-Pose`_ dataset.

    .. _5 Categories Animal-Pose: https://sites.google.com/view/animal-pose/

    The file structure should be like::

        <path>
            keypoint_image_part1/
                cat/
                    2007_000549.jpg
                    2007_000876.jpg
                    ...
                ...
            PASCAL2011_animal_annotation/
                cat/
                    2007_000549_1.xml
                    2007_000876_1.xml
                    2007_000876_2.xml
                    ...
                ...
            animalpose_image_part2/
                cat/
                    ca1.jpeg
                    ca2.jpeg
                    ...
                ...
            animalpose_anno2/
                cat/
                    ca1.xml
                    ca2.xml
                ...

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class:`~tensorbay.dataset.dataset.Dataset` instance.

    """
    root_path = os.path.abspath(os.path.expanduser(path))

    dataset = Dataset(DATASET_NAME_5)
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog_5.json"))
    animals = dataset.catalog.keypoints2d.categories.keys()

    for segment_name, _get_data in _DATA_GETTERS.items():
        segment = dataset.create_segment(segment_name)
        for data in _get_data(root_path, animals):
            segment.append(data)

    return dataset


def _get_data_part1(root_path: str, aniamls: Iterable[str]) -> Iterator[Data]:
    try:
        import xmltodict  # pylint: disable=import-outside-toplevel
    except ModuleNotFoundError as error:
        raise ModuleImportError(error.name) from error  # type: ignore[arg-type]

    for animal in aniamls:
        for image_path in glob(os.path.join(root_path, "keypoint_image_part1", animal, "*.jpg")):
            data = Data(image_path, target_remote_path=f"{animal}/{os.path.basename(image_path)}")

            for annotation_path in glob(
                os.path.join(
                    root_path,
                    "PASCAL2011_animal_annotation",
                    animal,
                    f"{os.path.splitext(os.path.basename(image_path))[0]}_*.xml",
                )
            ):

                with open(annotation_path, encoding="utf-8") as fp:
                    labels = xmltodict.parse(fp.read())

                box2d = labels["annotation"]["visible_bounds"]
                data.label.box2d = [
                    LabeledBox2D.from_xywh(
                        x=float(box2d["@xmin"]),
                        y=float(box2d["@ymin"]),
                        width=float(box2d["@width"]),
                        height=float(box2d["@height"]),
                        category=animal,
                    )
                ]

                keypoints2d: List[Tuple[float, float, int]] = [()] * 20  # type: ignore[list-item]
                for keypoint in labels["annotation"]["keypoints"]["keypoint"]:
                    keypoints2d[_KEYPOINT_TO_INDEX[keypoint["@name"]]] = (
                        float(keypoint["@x"]),
                        float(keypoint["@y"]),
                        int(keypoint["@visible"]),
                    )
                data.label.keypoints2d = [LabeledKeypoints2D(keypoints2d, category=animal)]

            yield data


def _get_data_part2(root_path: str, aniamls: Iterable[str]) -> Iterator[Data]:
    try:
        import xmltodict  # pylint: disable=import-outside-toplevel
    except ModuleNotFoundError as error:
        raise ModuleImportError(error.name) from error  # type: ignore[arg-type]

    for animal in aniamls:
        for image_path in glob(os.path.join(root_path, "animalpose_image_part2", animal, "*.jpeg")):
            data = Data(image_path, target_remote_path=f"{animal}/{os.path.basename(image_path)}")

            annotation_path = os.path.join(
                root_path,
                "animalpose_anno2",
                animal,
                f"{os.path.splitext(os.path.basename(image_path))[0]}.xml",
            )

            with open(annotation_path, encoding="utf-8") as fp:
                labels = xmltodict.parse(fp.read())

            box2d = labels["annotation"]["visible_bounds"]
            data.label.box2d = [
                LabeledBox2D.from_xywh(
                    x=float(box2d["@xmin"]),
                    y=float(box2d["@xmax"]),  # xmax means ymin in the annotation
                    width=float(box2d["@width"]),
                    height=float(box2d["@height"]),
                    category=animal,
                )
            ]

            keypoints2d = LabeledKeypoints2D(category=animal)
            for keypoint in labels["annotation"]["keypoints"]["keypoint"]:
                keypoints2d.append(  # pylint: disable=no-member
                    Keypoint2D(
                        float(keypoint["@x"]), float(keypoint["@y"]), int(keypoint["@visible"])
                    )
                )
            data.label.keypoints2d = [keypoints2d]
            yield data


_DATA_GETTERS = {"part1": _get_data_part1, "part2": _get_data_part2}


def AnimalPose7(path: str) -> Dataset:
    """Dataloader of `7 Categories Animal-Pose`_ dataset.

    .. _7 Categories Animal-Pose: https://sites.google.com/view/animal-pose/

    The file structure should be like::

        <path>
            bndbox_image/
                antelope/
                    Img-77.jpg
                    ...
                ...
            bndbox_anno/
                antelope.json
                ...

    Arguments:
        path: The root directory of the dataset.

    Returns:
        loaded :class:`~tensorbay.dataset.dataset.Dataset` object.

    """
    root_path = os.path.abspath(os.path.expanduser(path))

    dataset = Dataset(DATASET_NAME_7)
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog_7.json"))

    segment = dataset.create_segment()

    for animal in dataset.catalog.box2d.categories.keys():
        with open(os.path.join(root_path, "bndbox_anno", f"{animal}.json"), encoding="utf-8") as fp:
            annotations = json.load(fp)
        for image_name, box2ds in annotations.items():
            image_path = os.path.join(root_path, "bndbox_image", animal, image_name)
            data = Data(image_path, target_remote_path=f"{animal}/{image_name}")
            data.label.box2d = []

            for box2d in box2ds:
                coordinates = box2d["bndbox"]
                data.label.box2d.append(
                    LabeledBox2D(
                        float(coordinates["xmin"]),
                        float(coordinates["ymin"]),
                        float(coordinates["xmax"]),
                        float(coordinates["ymax"]),
                        category=animal,
                    )
                )

            segment.append(data)

    return dataset



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Exceptions about open data loader.

The content in this module is deprecated since v1.3.0, and will be removed in v1.5.0.

Please use :class:`~tensorbay.exception.NoFileError` instead of :class:`OpenDatasetNoFileError`.

"""

from ...exception import NoFileError

OpenDatasetNoFileError = NoFileError



#!/usr/bin/env python3
#
# Copytright 2020 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name, missing-module-docstring

import csv
import os
from collections import defaultdict
from typing import Dict, List

from ...dataset import Data, Dataset
from ...label import Box2DSubcatalog, Classification, ClassificationSubcatalog, LabeledBox2D
from .._utility import glob

_DATASET_NAME = "UAVDT"

_MODES = ("train", "test")

_FIELDNAMES = (
    "frame_index",
    "target_id",
    "bbox_left",
    "bbox_top",
    "bbox_width",
    "bbox_height",
    "out_of_view",
    "occlusion",
    "object_category",
)

_XYWH_KEYS = _FIELDNAMES[2:6]


def UAVDT(path: str) -> Dataset:
    """Dataloader of the `UAVDT`_ Dataset.

    .. _UAVDT: https://sites.google.com/site/daviddo0323/projects/uavdt

    The "score", "in-view", "occlusion" fields in MOT Groundtruth file(``*_gt.txt``) are constant,
    and other fields in that file are the same with such fields in DET Groundtruth file
    (``*_gt_whole.txt``). Therefore, they are not included in the dataloader.

    The Ignore Areas file(``*_gt_ignore.txt``) is useless,
    so they are not included in the dataloader neither.

    The file structure of UAVDT looks like::

        <path>
            M_attr/
                test/
                    M0203_attr.txt
                    ...
                train/
                    M0101_attr.txt
                    ...
            UAVDT_Benchmark_M/
                M0101/
                    img000001.jpg
                    ...
                ...
            UAV-benchmark-MOTD_v1.0/
                GT/
                    M0101_gt_ignore.txt
                    M0101_gt.txt
                    M0101_gt_whole.txt
                    ...

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class:`~tensorbay.dataset.dataset.Dataset` instance.

    """
    root_path = os.path.abspath(os.path.expanduser(path))

    dataset = Dataset(_DATASET_NAME)
    dataset.notes.is_continuous = True
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog.json"))

    for mode in _MODES:
        for sequence_attributes_path in glob(os.path.join(root_path, "M_attr", mode, "*.txt")):
            sequence = os.path.basename(sequence_attributes_path)[:5]
            segment = dataset.create_segment(f"{mode}-{sequence}")

            classification = _extract_classification(
                sequence_attributes_path, dataset.catalog.classification
            )

            frame_id_box2d_map = _extract_box2d(root_path, sequence, dataset.catalog.box2d)

            image_paths = glob(os.path.join(root_path, "UAV-benchmark-M", sequence, "*.jpg"))
            for image_path in image_paths:
                data = Data(image_path)
                data.label.classification = classification
                # The image_name looks like `img<frame_id>.jpg`.
                # The frame_id is consist of six digital numbers.
                data.label.box2d = frame_id_box2d_map[int(os.path.basename(image_path)[3:9])]
                segment.append(data)

    return dataset


def _extract_classification(
    path: str, classification_catalog: ClassificationSubcatalog
) -> Classification:
    with open(path) as fp:
        attribute_names = classification_catalog.attributes.keys()
        csv_reader = csv.reader(fp)
        elements = next(csv_reader)
        attributes = {
            attribute_name: bool(int(value))
            for attribute_name, value in zip(attribute_names, elements)
        }

    return Classification(attributes=attributes)


def _extract_box2d(
    path: str, sequence: str, box2d_catalog: Box2DSubcatalog
) -> Dict[int, List[LabeledBox2D]]:
    attributes = box2d_catalog.attributes
    category_names = box2d_catalog.categories.keys()

    out_of_view_level = attributes["out_of_view"].enum
    occlusion_level = attributes["occlusion"].enum

    ground_truth_path = os.path.join(path, "UAV-benchmark-MOTD_v1.0", "GT")
    frame_id_ground_truth_map = defaultdict(list)
    with open(os.path.join(ground_truth_path, f"{sequence}_gt_whole.txt")) as fp:
        for elements in csv.DictReader(fp, fieldnames=_FIELDNAMES):
            box2d = LabeledBox2D.from_xywh(
                *(int(elements[key]) for key in _XYWH_KEYS),
                category=category_names[int(elements["object_category"]) - 1],
                attributes={
                    "out_of_view": out_of_view_level[int(elements["out_of_view"]) - 1],
                    "occlusion": occlusion_level[int(elements["occlusion"]) - 1],
                },
                instance=elements["target_id"],
            )
            frame_id = int(elements["frame_index"])
            frame_id_ground_truth_map[frame_id].append(box2d)

    return frame_id_ground_truth_map



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""TensorBay cutoms exceptions.

The class hierarchy for TensorBay custom exceptions is::

     +-- TensorBayException
         +-- ClientError
             +-- StatusError
             +-- DatasetTypeError
             +-- FrameError
             +-- ResponseError
                 +-- AccessDeniedError
                 +-- InvalidParamsError
                 +-- NameConflictError
                 +-- RequestParamsMissingError
                 +-- ResourceNotExistError
                 +-- ResponseSystemError
                 +-- UnauthorizedError
        +-- UtilityError
            +-- AttrError
         +-- TBRNError
         +-- OpenDatasetError
             +-- NoFileError
             +-- FileStructureError

"""

from typing import Dict, Optional, Type, Union

from requests.models import Response


class TensorBayException(Exception):
    """This is the base class for TensorBay custom exceptions."""


class ClientError(TensorBayException):
    """This is the base class for custom exceptions in TensorBay client module."""


class StatusError(ClientError):
    """This class defines the exception for illegal status.

    Arguments:
        is_draft: Whether the status is draft.
        message: The error message.

    """

    def __init__(self, is_draft: Optional[bool] = None, message: Optional[str] = None) -> None:
        super().__init__()
        if is_draft is None:
            self._message = message
        else:
            required_status = "commit" if is_draft else "draft"
            self._message = f"The status is not {required_status}"

    def __str__(self) -> str:
        return self._message if self._message else ""


class DatasetTypeError(ClientError):
    """This class defines the exception for incorrect type of the requested dataset.

    Arguments:
        dataset_name: The name of the dataset whose requested type is wrong.
        is_fusion: Whether the dataset is a fusion dataset.

    """

    def __init__(self, dataset_name: str, is_fusion: bool) -> None:
        super().__init__()
        self._dataset_name = dataset_name
        self._is_fusion = is_fusion

    def __str__(self) -> str:
        return (
            f"Dataset '{self._dataset_name}' is {'' if self._is_fusion else 'not '}a fusion dataset"
        )


class FrameError(ClientError):
    """This class defines the exception for incorrect frame id.

    Arguments:
       message: The error message.

    """

    def __init__(self, message: str) -> None:
        super().__init__()
        self._message = message

    def __str__(self) -> str:
        return self._message


class OperationError(ClientError):
    """This class defines the exception for incorrect operation.

    Arguments:
       message: The error message.

    """

    def __init__(self, message: str) -> None:
        super().__init__()
        self._message = message

    def __str__(self) -> str:
        return self._message


class ResponseError(ClientError):
    """This class defines the exception for post response error.

    Arguments:
        response: The response of the request.

    Attributes:
        response: The response of the request.

    """

    # https://github.com/python/mypy/issues/6473
    _INDENT = " " * len(__qualname__)  # type: ignore[name-defined]

    STATUS_CODE: int

    def __init__(self, response: Response) -> None:
        super().__init__()
        self.response = response

    def __init_subclass__(cls) -> None:
        cls._INDENT = " " * len(cls.__name__)

    def __str__(self) -> str:
        return (
            f"Unexpected status code({self.response.status_code})! {self.response.url}!"
            f"\n{self._INDENT}  {self.response.text}"
        )


class AccessDeniedError(ResponseError):
    """This class defines the exception for access denied response error."""

    STATUS_CODE = 403


class InvalidParamsError(ResponseError):
    """This class defines the exception for invalid parameters response error.

    Arguments:
        response: The response of the request.
        param_name: The name of the invalid parameter.
        param_value: The value of the invalid parameter.

    Attributes:
        response: The response of the request.

    """

    STATUS_CODE = 400

    def __init__(  # pylint: disable=super-init-not-called
        self,
        response: Optional[Response] = None,
        *,
        param_name: Optional[str] = None,
        param_value: Optional[str] = None,
    ) -> None:
        if response is not None:
            super().__init__(response)
            return

        self._param_name = param_name
        self._param_value = param_value

    def __str__(self) -> str:
        if hasattr(self, "response"):
            return super().__str__()

        messages = [f"Invalid {self._param_name}: {self._param_value}."]
        if self._param_name == "path":
            messages.append("Remote path should follow linux style.")

        return f"\n{self._INDENT}".join(messages)


class NameConflictError(ResponseError):
    """This class defines the exception for name conflict response error.

    Arguments:
        response: The response of the request.
        resource: The type of the conflict resource.
        identification: The identification of the conflict resource.

    Attributes:
        response: The response of the request.

    """

    STATUS_CODE = 400

    def __init__(  # pylint: disable=super-init-not-called
        self,
        response: Optional[Response] = None,
        *,
        resource: Optional[str] = None,
        identification: Union[int, str, None] = None,
    ) -> None:
        if response is not None:
            super().__init__(response)
            return

        self._resource = resource
        self._identification = identification

    def __str__(self) -> str:
        if hasattr(self, "response"):
            return super().__str__()

        return f"The {self._resource}: {self._identification} already exists."


class RequestParamsMissingError(ResponseError):
    """This class defines the exception for request parameters missing response error."""

    STATUS_CODE = 400


class ResourceNotExistError(ResponseError):
    """This class defines the exception for resource not existing response error.

    Arguments:
        response: The response of the request.
        resource: The type of the conflict resource.
        identification: The identification of the conflict resource.

    Arguments:
        response: The response of the request.

    """

    STATUS_CODE = 404

    def __init__(  # pylint: disable=super-init-not-called
        self,
        response: Optional[Response] = None,
        *,
        resource: Optional[str] = None,
        identification: Union[int, str, None] = None,
    ) -> None:
        if response is not None:
            super().__init__(response)
            return

        self._resource = resource
        self._identification = identification

    def __str__(self) -> str:
        if hasattr(self, "response"):
            return super().__str__()

        return f"The {self._resource}: {self._identification} does not exist."


class ResponseSystemError(ResponseError):
    """This class defines the exception for system response error."""

    STATUS_CODE = 500


class UnauthorizedError(ResponseError):
    """This class defines the exception for unauthorized response error."""

    STATUS_CODE = 401


class OpenDatasetError(TensorBayException):
    """This is the base class for custom exceptions in TensorBay opendataset module."""


class NoFileError(OpenDatasetError):
    """This class defines the exception for no matching file found in the opendataset directory.

    Arguments:
        pattern: Glob pattern.

    """

    def __init__(self, pattern: str) -> None:
        super().__init__()
        self._pattern = pattern

    def __str__(self) -> str:
        return f'No file follows the giving pattern "{self._pattern}"'


class FileStructureError(OpenDatasetError):
    """This class defines the exception for incorrect file structure in the opendataset directory.

    Arguments:
        message: The error message.

    """

    def __init__(self, message: str) -> None:
        super().__init__()
        self._message = message

    def __str__(self) -> str:
        return self._message


class ModuleImportError(OpenDatasetError, ModuleNotFoundError):
    """This class defines the exception for import error of optional module in opendataset module.

    Arguments:
        module_name: The name of the optional module.
        package_name: The package name of the optional module.

    """

    def __init__(self, module_name: str, package_name: Optional[str] = None) -> None:
        super().__init__()
        self._module_name = module_name
        self._package_name = package_name if package_name else module_name

    def __str__(self) -> str:
        return (
            f"No module named {self._module_name}."
            "\n"
            f'\n    To install the module, please run: "pip3 install {self._package_name}"'
            "\n"
        )


class TBRNError(TensorBayException):
    """This class defines the exception for invalid TBRN.

    Arguments:
        message: The error message.

    """

    def __init__(self, message: str) -> None:
        super().__init__()
        self._message = message

    def __str__(self) -> str:
        return self._message


ResponseErrorDistributor: Dict[str, Type[ResponseError]] = {
    "AccessDenied": AccessDeniedError,
    "InvalidParamsValue": InvalidParamsError,
    "NameConflict": NameConflictError,
    "RequestParamsMissing": RequestParamsMissingError,
    "ResourceNotExist": ResourceNotExistError,
    "SystemError": ResponseSystemError,
    "Unauthorized": UnauthorizedError,
}


class UtilityError(TensorBayException):
    """This is the base class for custom exceptions in TensorBay utility module."""


class AttrError(UtilityError):
    """This class defines the exception for dynamic attr have default value."""

    def __str__(self) -> str:
        return "Dynamic attr cannot have default value."



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Utility classes."""

from .attr import AttrsMixin, attr, attr_base, camel, upper
from .common import EqMixin, MatrixType, common_loads, locked
from .deprecated import DefaultValueDeprecated, Deprecated, Disable, KwargsDeprecated
from .file import FileMixin, RemoteFileMixin
from .itertools import chunked
from .name import NameList, NameMixin, SortedNameList
from .repr import ReprMixin, ReprType, repr_config
from .type import TypeEnum, TypeMixin, TypeRegister
from .user import UserMapping, UserMutableMapping, UserMutableSequence, UserSequence

__all__ = [
    "AttrsMixin",
    "DefaultValueDeprecated",
    "Deprecated",
    "Disable",
    "EqMixin",
    "FileMixin",
    "KwargsDeprecated",
    "MatrixType",
    "NameList",
    "NameMixin",
    "RemoteFileMixin",
    "ReprMixin",
    "ReprType",
    "SortedNameList",
    "TypeEnum",
    "TypeMixin",
    "TypeRegister",
    "UserMapping",
    "UserMutableMapping",
    "UserMutableSequence",
    "UserSequence",
    "attr",
    "attr_base",
    "camel",
    "chunked",
    "common_loads",
    "locked",
    "repr_config",
    "upper",
]



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name, missing-module-docstring

import os
from xml.etree import ElementTree

from ...dataset import Data, Dataset
from ...label import LabeledBox2D

_SEGMENT_NAMES = (
    "train",
    "trainval",
    "val",
)
_BOOLEAN_ATTRIBUTES = {"occluded", "pose", "truncated"}
DATASET_NAME = "VOC2012Detection"


def VOC2012Detection(path: str) -> Dataset:
    """Dataloader of the 'VOC2012Detection'_ dataset.

    .. _VOC2012Detection: http://host.robots.ox.ac.uk/pascal/VOC/voc2012/

    The file structure should be like::

        <path>
            Annotations/
                <image_name>.xml
                ...
            JPEGImages/
                <image_name>.jpg
                ...
            ImageSets/
                Main/
                    train.txt
                    trainval.txt
                    val.txt
                    ...
                ...
            ...

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class: `~tensorbay.dataset.dataset.Dataset` instance.

    """
    annotation_path = os.path.join(path, "Annotations")
    image_path = os.path.join(path, "JPEGImages")
    main_path = os.path.join(path, "ImageSets", "Main")

    dataset = Dataset(DATASET_NAME)
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog.json"))

    for segment_name in _SEGMENT_NAMES:
        segment = dataset.create_segment(segment_name)
        with open(os.path.join(main_path, f"{segment_name}.txt")) as fp:
            for filename in fp:
                filename = filename.strip()
                segment.append(_get_data(filename, image_path, annotation_path))
    return dataset


def _get_data(filename: str, image_path: str, annotation_path: str) -> Data:
    """Get all information of the datum corresponding to filename.

    Arguments:
        filename: The filename of the data.
        image_path: The path of the image directory.
        annotation_path: The path of the annotation directory.

    Returns:
        Data: class: `~tensorbay.dataset.data.Data` instance.

    """
    data = Data(os.path.join(image_path, f"{filename}.jpg"))
    data.label.box2d = []
    tree = ElementTree.parse(os.path.join(annotation_path, f"{filename}.xml"))
    for obj in tree.findall("object"):
        attributes = {}
        for child in obj:
            if child.tag == "name":
                category = child.text
            elif child.tag == "bndbox":
                box = (
                    float(child.find("xmin").text),  # type:ignore[arg-type, union-attr]
                    float(child.find("ymin").text),  # type:ignore[arg-type, union-attr]
                    float(child.find("xmax").text),  # type:ignore[arg-type, union-attr]
                    float(child.find("ymax").text),  # type:ignore[arg-type, union-attr]
                )
            elif child.tag == "pose":
                attributes[child.tag] = child.text
            elif child.tag in _BOOLEAN_ATTRIBUTES:
                attributes[child.tag] = bool(
                    int(child.text)  # type:ignore[assignment, arg-type]
                )
        data.label.box2d.append(LabeledBox2D(*box, category=category, attributes=attributes))
    return data



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name
# pylint: disable=missing-module-docstring


import json
import os
from typing import Any, Dict, Iterator, List

from ...dataset import Data, Frame, FusionDataset, FusionSegment
from ...geometry import Transform3D
from ...label import LabeledBox3D
from ...sensor import Camera, Lidar, Radar

DATASET_NAME = "nuScenes"

_SENSOR_TYPE_CLASS = {
    "lidar": Lidar,
    "radar": Radar,
    "camera": Camera,
}
_ATTRIBUTE_KEYS = {
    "vehicle": "vehicle_motion",
    "cycle": "cycle_rider",
    "pedestrian": "pedestrian_motion",
}


def nuScenes(path: str) -> FusionDataset:
    """Dataloader of the `nuScenes`_ dataset.

    .. _nuScenes: https://www.nuscenes.org/

    The file structure should be like::

        <path>
            v1.0-mini/
                maps/
                    36092f0b03a857c6a3403e25b4b7aab3.png
                    ...
                samples/
                    CAM_BACK/
                    CAM_BACK_LEFT/
                    CAM_BACK_RIGHT/
                    CAM_FRONT/
                    CAM_FRONT_LEFT/
                    CAM_FRONT_RIGHT/
                    LIDAR_TOP/
                    RADAR_BACK_LEFT/
                    RADAR_BACK_RIGHT/
                    RADAR_FRONT/
                    RADAR_FRONT_LEFT/
                    RADAR_FRONT_RIGHT/
                sweeps/
                    CAM_BACK/
                    CAM_BACK_LEFT/
                    CAM_BACK_RIGHT/
                    CAM_FRONT/
                    CAM_FRONT_LEFT/
                    CAM_FRONT_RIGHT/
                    LIDAR_TOP/
                    RADAR_BACK_LEFT/
                    RADAR_BACK_RIGHT/
                    RADAR_FRONT/
                    RADAR_FRONT_LEFT/
                    RADAR_FRONT_RIGHT/
                v1.0-mini/
                    attribute.json
                    calibrated_sensor.json
                    category.json
                    ego_pose.json
                    instance.json
                    log.json
                    map.json
                    sample_annotation.json
                    sample_data.json
                    sample.json
                    scene.json
                    sensor.json
                    visibility.json
            v1.0-test/
                maps/
                samples/
                sweeps/
                v1.0-test/
            v1.0-trainval/
                maps/
                samples/
                sweeps/
                v1.0-trainval/

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class:`~tensorbay.dataset.dataset.FusionDataset` instance.

    """
    root_path = os.path.abspath(os.path.expanduser(path))

    dataset = FusionDataset(DATASET_NAME)
    dataset.notes.is_continuous = True
    dataset.notes.bin_point_cloud_fields = ["X", "Y", "Z", "Intensity", "Ring"]
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog.json"))

    for subset in os.listdir(root_path):
        for segment in _generate_segments(root_path, subset):
            dataset.add_segment(segment)

    return dataset


def _generate_segments(root_path: str, subset: str) -> Iterator[FusionSegment]:
    subset_path = os.path.join(root_path, subset)
    is_test = subset.endswith("test")
    annotation_info = _get_annotation_info(os.path.join(subset_path, subset), is_test)

    for scene in annotation_info["scenes"]:
        segment = FusionSegment(f"{subset}-{scene['name']}")
        segment.description = scene["description"]
        current_frame_token = scene["first_sample_token"]
        while current_frame_token:
            _load_frame_and_sensor(
                segment,
                current_frame_token,
                annotation_info,
                subset_path,
                is_test,
            )
            current_frame_token = annotation_info["samples"][current_frame_token]["next"]
        yield segment


def _get_annotation_info(info_path: str, is_test: bool = False) -> Dict[str, Any]:
    annotation_info = {
        "samples": _get_info_with_token(info_path, "sample"),
        "frame_data": _get_info_with_sample_token(info_path, "sample_data"),
        "calibrated_sensors": _get_info_with_token(info_path, "calibrated_sensor"),
        "ego_poses": _get_info_with_token(info_path, "ego_pose"),
        "sensor": _get_info_with_token(info_path, "sensor"),
    }
    with open(os.path.join(info_path, "scene.json"), "r") as file:
        annotation_info["scenes"] = json.load(file)
    if not is_test:
        annotation_info["sample_annotations"] = _get_info_with_sample_token(
            info_path, "sample_annotation", no_key_frame=True
        )
        annotation_info["instance"] = _get_info_with_token(info_path, "instance")
        annotation_info["category"] = _get_info_with_token(info_path, "category")
        annotation_info["attribute"] = _get_info_with_token(info_path, "attribute")
        annotation_info["visibility"] = _get_info_with_token(info_path, "visibility")

    return annotation_info


def _get_info_with_token(info_path: str, annotation_part: str) -> Dict[str, Any]:
    filepath = os.path.join(info_path, f"{annotation_part}.json")
    with open(filepath, "r") as fp:
        info = json.load(fp)
    return {item.pop("token"): item for item in info}


def _get_info_with_sample_token(
    info_path: str, annotation_part: str, no_key_frame: bool = False
) -> Dict[str, List[Any]]:
    filepath = os.path.join(info_path, f"{annotation_part}.json")
    with open(filepath, "r") as fp:
        info = json.load(fp)
    info_with_keys: Dict[str, List[Any]] = {}
    for item in info:
        if no_key_frame or item["is_key_frame"]:
            info_with_keys.setdefault(item.pop("sample_token"), []).append(item)
    return info_with_keys


def _load_frame_and_sensor(
    segment: FusionSegment,
    current_frame_token: str,
    annotation_info: Dict[str, Any],
    subset_path: str,
    is_test: bool,
) -> None:
    frame = Frame()
    for sensor_frame in annotation_info["frame_data"][current_frame_token]:
        calibrated_sensor_info = annotation_info["calibrated_sensors"][
            sensor_frame["calibrated_sensor_token"]
        ]
        common_sensor = annotation_info["sensor"][calibrated_sensor_info["sensor_token"]]
        sensor_name = common_sensor["channel"]
        sensor_type = common_sensor["modality"]

        if sensor_name not in segment.sensors:
            _load_sensor(segment, sensor_type, sensor_name, calibrated_sensor_info)

        data = Data(
            os.path.join(subset_path, sensor_frame["filename"]),
            timestamp=sensor_frame["timestamp"] / 10 ** 6,
        )

        if not is_test and sensor_type == "lidar":
            _load_labels(
                data,
                current_frame_token,
                annotation_info["ego_poses"][sensor_frame["ego_pose_token"]],
                segment.sensors[sensor_name].extrinsics,
                annotation_info,
            )

        frame[sensor_name] = data
    segment.append(frame)


def _load_sensor(
    segment: FusionSegment,
    sensor_type: str,
    sensor_name: str,
    calibrated_sensor_info: Dict[str, Any],
) -> None:
    sensor = _SENSOR_TYPE_CLASS[sensor_type](sensor_name)
    sensor.set_extrinsics(
        translation=calibrated_sensor_info["translation"],
        rotation=calibrated_sensor_info["rotation"],
    )
    intrinsics = calibrated_sensor_info["camera_intrinsic"]
    if intrinsics:
        sensor.set_camera_matrix(matrix=intrinsics)  # type: ignore[attr-defined]
    segment.sensors.add(sensor)  # type: ignore[arg-type]


def _load_labels(
    data: Data,
    current_frame_token: str,
    lidar_ego_pose_info: Dict[str, Any],
    lidar_to_ego: Transform3D,
    annotation_info: Dict[str, Dict[str, Any]],
) -> None:
    labels = []
    sample_annotations = annotation_info["sample_annotations"]
    if current_frame_token in sample_annotations:
        lidar_ego_pose = Transform3D(
            translation=lidar_ego_pose_info["translation"],
            rotation=lidar_ego_pose_info["rotation"],
        )
        world_to_lidar = (lidar_ego_pose * lidar_to_ego).inverse()
        for instance_annotation in sample_annotations[current_frame_token]:
            labeled_box = _get_labeled_box(instance_annotation, annotation_info)
            labels.append(world_to_lidar * labeled_box)
    data.label.box3d = labels


def _get_labeled_box(
    instance_annotation: Dict[str, Any],
    annotation_info: Dict[str, Any],
) -> LabeledBox3D:
    instance = instance_annotation["instance_token"]
    category = annotation_info["category"][annotation_info["instance"][instance]["category_token"]][
        "name"
    ]
    attributes = {}
    for attribute_token in instance_annotation["attribute_tokens"]:
        attribute_full_name = annotation_info["attribute"][attribute_token]["name"]
        key, value = attribute_full_name.rsplit(".", 1)
        attributes[_ATTRIBUTE_KEYS[key]] = value
    attributes["visibility"] = annotation_info["visibility"][
        instance_annotation["visibility_token"]
    ]["level"]

    width, length, height = instance_annotation["size"]
    return LabeledBox3D(
        translation=instance_annotation["translation"],
        rotation=instance_annotation["rotation"],
        size=(length, width, height),
        category=category,
        instance=instance,
        attributes=attributes,
    )



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Implementation of gas cp."""

import os
from pathlib import Path, PurePosixPath
from typing import Dict, Iterable

from ..dataset import Data, Segment
from .tbrn import TBRN, TBRNType
from .utility import error, get_dataset_client, get_gas


def _implement_cp(  # pylint: disable=too-many-arguments
    obj: Dict[str, str],
    local_paths: Iterable[str],
    tbrn: str,
    is_recursive: bool,
    jobs: int,
    skip_uploaded_files: bool,
) -> None:
    gas = get_gas(**obj)
    info = TBRN(tbrn=tbrn)

    dataset_client = get_dataset_client(gas, info, is_fusion=False)

    if info.type not in (TBRNType.SEGMENT, TBRNType.NORMAL_FILE):
        error(f'"{tbrn}" is not a segment or file type')

    target_remote_path = info.remote_path if info.type == TBRNType.NORMAL_FILE else ""

    local_abspaths = [os.path.abspath(local_path) for local_path in local_paths]
    if (
        len(local_abspaths) == 1
        and not os.path.isdir(local_abspaths[0])
        and target_remote_path
        and not target_remote_path.endswith("/")
    ):
        segment_client = dataset_client.get_or_create_segment(info.segment_name)
        segment_client.upload_file(local_abspaths[0], target_remote_path)
    else:
        segment = _get_segment(info.segment_name, local_abspaths, target_remote_path, is_recursive)
        dataset_client.upload_segment(segment, jobs=jobs, skip_uploaded_files=skip_uploaded_files)


def _get_segment(
    segment_name: str,
    local_abspaths: Iterable[str],
    remote_path: str,
    is_recursive: bool,
) -> Segment:
    """Get the pair of local_path and remote_path.

    Arguments:
        segment_name: The name of the segment these data belong to.
        local_abspaths: A list of local abstract paths, could be folder or file.
        remote_path: The remote object path, not necessarily end with '/'.
        is_recursive: Whether copy directories recursively.

    Returns:
        A segment contains mapping data.

    """
    segment = Segment(segment_name)
    for local_abspath in local_abspaths:
        if not os.path.isdir(local_abspath):
            data = Data(
                local_abspath,
                target_remote_path=str(PurePosixPath(remote_path, os.path.basename(local_abspath))),
            )
            segment.append(data)
            continue

        if not is_recursive:
            error("Local paths include directories, please use -r option")

        local_abspath = os.path.normpath(local_abspath)
        folder_name = os.path.basename(local_abspath)
        for root, _, filenames in os.walk(local_abspath):
            relpath = os.path.relpath(root, local_abspath) if root != local_abspath else ""
            for filename in filenames:
                data = Data(
                    os.path.join(root, filename),
                    target_remote_path=str(
                        PurePosixPath(Path(remote_path, folder_name, relpath, filename))
                    ),
                )
                segment.append(data)
    return segment



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""PointList2D, MultiPointList2D.

:class:`PointList2D` contains a list of 2D points.

:class:`MultiPointList2D` contains multiple 2D point lists.

"""

from typing import Any, Dict, Iterable, List, Optional, Type, TypeVar

from ..utility import UserMutableSequence, common_loads
from .box import Box2D
from .vector import Vector2D

_T = TypeVar("_T", bound=Vector2D)


class PointList2D(UserMutableSequence[_T]):  # pylint: disable=too-many-ancestors
    """This class defines the concept of PointList2D.

    :class:`PointList2D` contains a list of 2D points.

    Arguments:
        points: A list of 2D points.

    """

    _P = TypeVar("_P", bound="PointList2D[_T]")

    _ElementType: Type[_T]

    def __init__(
        self,
        points: Optional[Iterable[Iterable[float]]] = None,
    ) -> None:
        self._data = [self._ElementType(*point) for point in points] if points else []

    def _loads(self, contents: List[Dict[str, float]]) -> None:
        self._data = []
        for point in contents:
            self._data.append(self._ElementType.loads(point))

    @classmethod
    def loads(cls: Type[_P], contents: List[Dict[str, float]]) -> _P:
        """Load a :class:`PointList2D` from a list of dictionaries.

        Arguments:
            contents: A list of dictionaries containing the coordinates of the vertexes
                of the point list::

                    [
                        {
                            "x": ...
                            "y": ...
                        },
                        ...
                    ]

        Returns:
            The loaded :class:`PointList2D` object.

        """
        return common_loads(cls, contents)

    def dumps(self) -> List[Dict[str, float]]:
        """Dumps a :class:`PointList2D` into a point list.

        Returns:
            A list of dictionaries containing the coordinates of the vertexes
            of the polygon within the point list.

        """
        return [point.dumps() for point in self._data]

    def bounds(self) -> Box2D:
        """Calculate the bounds of point list.

        Returns:
            The bounds of point list.

        """
        x_min = x_max = self._data[0].x
        y_min = y_max = self._data[0].y

        for point in self._data:
            if point.x < x_min:
                x_min = point.x
            elif point.x > x_max:
                x_max = point.x

            if point.y < y_min:
                y_min = point.y
            elif point.y > y_max:
                y_max = point.y

        return Box2D(x_min, y_min, x_max, y_max)


_L = TypeVar("_L", bound=PointList2D[Any])


class MultiPointList2D(UserMutableSequence[_L]):  # pylint: disable=too-many-ancestors
    """This class defines the concept of MultiPointList2D.

    :class:`MultiPointList2D` contains multiple 2D point lists.

    Arguments:
        point_lists: A list of 2D point list.

    """

    _P = TypeVar("_P", bound="MultiPointList2D[_L]")

    _ElementType: Type[_L]

    def __init__(self, point_lists: Optional[Iterable[Iterable[Iterable[float]]]] = None) -> None:
        self._data = (
            [self._ElementType(point_list) for point_list in point_lists] if point_lists else []
        )

    def _loads(self, contents: List[List[Dict[str, float]]]) -> None:
        self._data = [self._ElementType.loads(point_list) for point_list in contents]

    def _dumps(self) -> List[List[Dict[str, float]]]:
        return [point_list.dumps() for point_list in self._data]

    @classmethod
    def loads(cls: Type[_P], contents: List[List[Dict[str, float]]]) -> _P:
        """Loads a :class:`MultiPointList2D` from the given contents.

        Arguments:
            contents: A list of dictionary lists containing the coordinates of the vertexes
                of the multiple point lists::

                    [
                        [
                            {
                                "x": ...
                                "y": ...
                            },
                            ...
                        ]
                        ...
                    ]

        Returns:
            The loaded :class:`MultiPointList2D` object.

        """
        return common_loads(cls, contents)

    def dumps(self) -> List[List[Dict[str, float]]]:
        """Dumps all the information of the :class:`MultiPointList2D`.

        Returns:
            All the information of the :class:`MultiPointList2D`.

        """
        return self._dumps()

    def bounds(self) -> Box2D:
        """Calculate the bounds of multiple point lists.

        Returns:
            The bounds of multiple point lists.

        """
        x_min = x_max = self._data[0][0].x
        y_min = y_max = self._data[0][0].y

        for points in self._data:
            box = points.bounds()
            if box.xmin < x_min:
                x_min = box.xmin
            elif box.xmax > x_max:
                x_max = box.xmax

            if box.ymin < y_min:
                y_min = box.ymin
            elif box.ymax > y_max:
                y_max = box.ymax

        return Box2D(x_min, y_min, x_max, y_max)



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name, missing-module-docstring

import os

from ...dataset import Data, Dataset
from ...exception import ModuleImportError
from ...label import LabeledBox2D

_SEGMENT_NAMES = (
    "train",
    "trainval",
    "val",
)
DATASET_NAME = "VOC2012ActionClassification"


def VOC2012ActionClassification(path: str) -> Dataset:
    """Dataloader of the 'VOC2012ActionClassification'_ dataset.

    .. _VOC2012ActionClassification: http://host.robots.ox.ac.uk/pascal/VOC/voc2012/

    The file structure should be like::

        <path>
            Annotations/
                <image_name>.xml
                ...
            JPEGImages/
                <image_name>.jpg
                ...
            ImageSets/
                Action/
                    train.txt
                    trainval.txt
                    val.txt
                    ...
                ...
            ...

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class: `~tensorbay.dataset.dataset.Dataset` instance.

    """
    annotation_path = os.path.join(path, "Annotations")
    image_path = os.path.join(path, "JPEGImages")
    action_path = os.path.join(path, "ImageSets", "Action")

    dataset = Dataset(DATASET_NAME)
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog.json"))

    for segment_name in _SEGMENT_NAMES:
        segment = dataset.create_segment(segment_name)
        with open(os.path.join(action_path, f"{segment_name}.txt")) as fp:
            for filename in fp:
                filename = filename.strip()
                segment.append(_get_data(filename, image_path, annotation_path))
    return dataset


def _get_data(filename: str, image_path: str, annotation_path: str) -> Data:
    try:
        import xmltodict  # pylint: disable=import-outside-toplevel
    except ModuleNotFoundError as error:
        raise ModuleImportError(error.name) from error  # type: ignore[arg-type]

    data = Data(os.path.join(image_path, f"{filename}.jpg"))
    box2d = []
    with open(os.path.join(annotation_path, f"{filename}.xml"), "r") as fp:
        objects = xmltodict.parse(fp.read())["annotation"]["object"]
    if not isinstance(objects, list):
        objects = [objects]
    for item in objects:
        category = item["name"]
        attributes = {k: bool(int(v)) for k, v in item["actions"].items()}
        bndbox = item["bndbox"]
        box2d.append(
            LabeledBox2D(
                float(bndbox["xmin"]),
                float(bndbox["ymin"]),
                float(bndbox["xmax"]),
                float(bndbox["ymax"]),
                category=category,
                attributes=attributes,
            )
        )
    data.label.box2d = box2d
    return data



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name
# pylint: disable=missing-module-docstring
# pylint: disable=line-too-long

import json
import os
from typing import Any, Dict

from ...dataset import Data, Dataset
from ...geometry import Keypoint2D
from ...label import LabeledKeypoints2D
from ...utility import chunked

DATASET_NAME = "HalpeFullBody"

_SEGMENT_SPLIT = (
    ("train", "halpe_train_v1.json", os.path.join("hico_20160224_det", "images", "train2015")),
    ("val", "halpe_val_v1.json", "val2017"),
)


def HalpeFullBody(path: str) -> Dataset:
    """Dataloader of the `Halpe Full-Body Human Keypoints and HOI-Det`_ dataset.

    .. _Halpe Full-Body Human Keypoints and HOI-Det: https://github.com/Fang-Haoshu/Halpe-FullBody/ # noqa: E501

    The folder structure should be like::

        <path>
            halpe_train_v1.json
            halpe_val_v1.json
            hico_20160224_det/
                images/
                    train2015/
                        HICO_train2015_00000001.jpg
                        ...
            val2017/
                000000000139.jpg
                ...

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class:`~tensorbay.dataset.dataset.Dataset` instance.

    """
    root_path = os.path.abspath(os.path.expanduser(path))

    dataset = Dataset(DATASET_NAME)
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog.json"))

    for mode, label_file, image_directory in _SEGMENT_SPLIT:
        segment = dataset.create_segment(mode)

        with open(os.path.join(root_path, label_file), "r", encoding="utf-8") as fp:
            annotations = json.load(fp)

        if mode == "train":
            for annotation, image in zip(annotations["annotations"], annotations["images"]):
                image_path = os.path.join(root_path, image_directory, image["file_name"])
                segment.append(_get_data(image_path, annotation))
        else:
            for annotation in annotations["annotations"]:
                image_path = os.path.join(
                    root_path, image_directory, f"{annotation['image_id']:012}.jpg"
                )
                segment.append(_get_data(image_path, annotation))

    return dataset


def _get_data(image_path: str, annotation: Dict[str, Any]) -> Data:
    data = Data(image_path)

    keypoints = LabeledKeypoints2D()
    for x, y, v in chunked(annotation["keypoints"], 3):
        keypoints.append(Keypoint2D(x, y, v if v in (0, 1, 2) else 2))  # pylint: disable=no-member

    data.label.keypoints2d = [keypoints]

    return data



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Graviti Tensorbay gas CLI utility functions."""

import logging
import os
import sys
from collections import OrderedDict
from configparser import ConfigParser, SectionProxy
from typing import Optional, Tuple, overload

import click
from typing_extensions import Literal, NoReturn

from ..client import GAS
from ..client import config as client_config
from ..client.dataset import DatasetClient, FusionDatasetClient
from ..client.gas import DatasetClientType
from .tbrn import TBRN


def _implement_cli(
    ctx: click.Context, access_key: str, url: str, profile_name: str, debug: bool
) -> None:
    ctx.obj = {
        "access_key": access_key,
        "url": url,
        "profile_name": profile_name,
    }

    if debug:
        logging.basicConfig(level=logging.DEBUG)


def _get_config_filepath() -> str:
    """Get the path of the config file.

    Returns:
        The path of the config file.

    """
    home = "USERPROFILE" if is_win() else "HOME"
    return os.path.join(os.environ[home], ".gasconfig")


def is_win() -> bool:
    """Determine whether the current system is Windows.

    Returns:
        Whether the current system is Windows.

    """
    return sys.platform.startswith("win")


def read_profile(profile_name: str, config_parser: ConfigParser) -> Tuple[str, str]:
    """Read accessKey and URL from the config file.

    Arguments:
        profile_name: The environment to login.
        config_parser: The config parser read from the config file.

    Returns:
        A tuple containing the accessKey and the url of profile_name read from the config file.

    """
    if not config_parser.has_option("profiles", profile_name):
        error(
            f"{profile_name} does not exist"
            f"\n\nPlease use 'gas -p {profile_name} auth [accessKey]' to create profile"
        )
    values = config_parser["profiles"][profile_name].split("\n")
    if len(values) == 2:
        return values[1], ""
    return values[1], values[2]


def get_gas(access_key: str, url: str, profile_name: str) -> GAS:
    """Load an object of :class:`~tensorbay.client.gas.GAS`.

    We will read accessKey and URL from the appointed profile_name and login gas.

    Arguments:
        access_key: The accessKey of gas.
        url: The login URL.
        profile_name: The environment to login.

    Returns:
        Gas client logged in with accessKey and URL.

    """
    update_config()
    config_parser = read_config()
    _set_request_config(config_parser)
    if not access_key and not url:
        access_key, url = read_profile(profile_name, config_parser)

    if not access_key:
        error("AccessKey should be appointed")

    return GAS(access_key, url)


@overload
def get_dataset_client(gas: GAS, info: TBRN, is_fusion: Literal[None] = None) -> DatasetClientType:
    ...


@overload
def get_dataset_client(gas: GAS, info: TBRN, is_fusion: Literal[False]) -> DatasetClient:
    ...


@overload
def get_dataset_client(gas: GAS, info: TBRN, is_fusion: Literal[True]) -> FusionDatasetClient:
    ...


@overload
def get_dataset_client(gas: GAS, info: TBRN, is_fusion: Optional[bool] = None) -> DatasetClientType:
    ...


def get_dataset_client(gas: GAS, info: TBRN, is_fusion: Optional[bool] = None) -> DatasetClientType:
    """Get the dataset client with any type and its version info.

    Arguments:
        gas: The gas client.
        info: The tbrn of the resource.
        is_fusion: Whether the dataset is a fusion dataset, True for fusion dataset.

    Returns:
        The dataset client and its version info.

    """
    dataset_client = (
        gas._get_dataset_with_any_type(info.dataset_name)  # pylint: disable=protected-access
        if is_fusion is None
        else gas.get_dataset(info.dataset_name, is_fusion)
    )
    if info.is_draft:
        dataset_client.checkout(draft_number=info.draft_number)
    elif info.revision is not None:
        dataset_client.checkout(revision=info.revision)
    return dataset_client


def _edit_input(hint: str) -> Tuple[str, str]:
    """Edit information input from the editor.

    Arguments:
        hint: The hint to be added in the temp file opened by the editor.

    Returns:
        The extracted title and the description.

    """
    config_parser = read_config()

    editor = config_parser["config"].get("editor") if config_parser.has_section("config") else None
    input_info = click.edit(hint, editor=editor, require_save=False)
    return _clean_up(input_info)


def _clean_up(editor_input: Optional[str]) -> Tuple[str, str]:
    """Clean up the information from the editor input.

    Arguments:
        editor_input: The editor input.

    Returns:
        The extracted title and the description.

    """
    if not editor_input:
        return "", ""

    cleaned_up_lines = []
    for line in editor_input.splitlines():
        if line and not line.startswith("#"):
            cleaned_up_lines.append(line.rstrip())

    if not cleaned_up_lines:
        return "", ""
    return cleaned_up_lines[0].lstrip(), "\n".join(cleaned_up_lines[1:])


def error(message: str) -> NoReturn:
    """Print the error message and exit the program.

    Arguments:
        message: The error message to echo.

    """
    click.secho(f"ERROR: {message}", err=True, fg="red")
    sys.exit(1)


def update_config() -> None:
    """Update the config file to the new format."""
    old_config_parser = read_config()
    old_sections = old_config_parser.sections()

    if not old_sections or old_sections == ["config"]:
        return

    if (old_sections in (["config", "profiles"], ["profiles"])) and _is_updated(
        old_config_parser["profiles"]
    ):
        return

    new_config_parser = ConfigParser(dict_type=OrderedDict)
    new_config_parser.add_section("profiles")
    for section_name, section_value in old_config_parser.items():
        if section_name == "DEFAULT":
            continue

        if section_name == "config":
            new_config_parser.add_section("config")
            new_config_parser["config"] = old_config_parser["config"]

        elif section_name == "profiles" and _is_updated(section_value):
            new_config_parser["profiles"].update(section_value)

        else:
            new_config_parser["profiles"][section_name] = form_profile_value(**section_value)

    write_config(new_config_parser, show_message=False)


def _is_updated(profile_section: SectionProxy) -> bool:
    return "\n" in profile_section.get("accesskey", "\n")


def form_profile_value(accesskey: str, url: Optional[str] = None) -> str:
    """Form the profile value with accesskey (and url).

    Arguments:
        accesskey: The accesskey to TensorBay.
        url: The TensorBay url.

    Returns:
        The formed profile value.

    """
    values = ["", accesskey]
    if url:
        values.append(url)
    return "\n".join(values)


def read_config(config_filepath: Optional[str] = None) -> ConfigParser:
    """Write the config from the config file.

    Arguments:
        config_filepath: The path of the config file to read.

    Returns:
        The config parser read from the config file.

    """
    if not config_filepath:
        config_filepath = _get_config_filepath()

    config_parser = ConfigParser(dict_type=OrderedDict)
    config_parser.read(config_filepath)
    return config_parser


def write_config(config_parser: ConfigParser, show_message: bool = True) -> None:
    """Write the config parser to the config file.

    Arguments:
        config_parser: The config parser to write to the file.
        show_message: Whether to show the message.

    """
    # pylint: disable=protected-access
    if config_parser.has_section("config"):
        config_parser._sections.move_to_end("config", last=False)  # type: ignore[attr-defined]
    if config_parser.has_section("profiles") and config_parser.has_option("profiles", "default"):
        config_parser._sections["profiles"].move_to_end(  # type: ignore[attr-defined]
            "default", last=False
        )

    config_file = _get_config_filepath()
    with open(config_file, "w") as fp:
        config_parser.write(fp)
    if show_message:
        click.echo(f'Success!\nConfiguration has been written into: "{config_file}"')


def is_accesskey(arg: str) -> bool:
    """Determine whether the string is an AccessKey.

    Arguments:
        arg: The string needed to be judged.

    Returns:
        Whether the string is an AccessKey.

    """
    return arg.startswith(("Accesskey-", "ACCESSKEY-")) and len(arg) == 42


def shorten(origin: str) -> str:
    """Return the first 7 characters of the original string.

    Arguments:
        origin: The string needed to be shortened.

    Returns:
        A string of length 7.

    """
    return origin[:7]


def format_hint(title: str, description: str, original_hint: str) -> str:
    """Generate complete hint message.

    Arguments:
        title: The title of the draft to edit or commit.
        description: The description of the draft to edit or commit.
        original_hint: The original hint message.

    Returns:
        The complete hint message.

    """
    hint: Tuple[str, ...] = (title,)
    if description:
        hint += ("", description)
    hint += (original_hint,)
    return "\n".join(hint)


def edit_message(message: Tuple[str, ...], hint_message: str) -> Tuple[str, str]:
    """Edit draft information.

    Arguments:
        message: The message given in the CLI.
        hint_message: The hint message to show on the pop-up editor.

    Returns:
        The extracted title and the description.

    """
    if message:
        title, description = message[0], "\n".join(message[1:])
    else:
        title, description = _edit_input(hint_message)

    return title, description


def _set_request_config(config_parser: ConfigParser) -> None:
    """Configure request related parameters.

    Arguments:
        config_parser: The config parser read from the config file.

    """
    client_config._x_source = "PYTHON-CLI"  # pylint: disable=protected-access
    if config_parser.has_section("config"):
        config_section = config_parser["config"]
        if "timeout" in config_section:
            client_config.timeout = config_section.getint("timeout")
        if "is_internal" in config_section:
            client_config.is_internal = config_section.getboolean("is_internal")
        if "max_retries" in config_section:
            client_config.max_retries = config_section.getint("max_retries")



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name
# pylint: disable=missing-module-docstring

import os
from typing import List
from xml.etree import ElementTree

from ...dataset import Data, Dataset
from ...label import LabeledBox2D
from .._utility import glob

DATASET_NAME = "HardHatWorkers"


def HardHatWorkers(path: str) -> Dataset:
    """Dataloader of the `Hard Hat Workers`_ dataset.

    .. _Hard Hat Workers: https://makeml.app/datasets/hard-hat-workers

    The file structure should be like::

        <path>
            annotations/
                hard_hat_workers0.xml
                ...
            images/
                hard_hat_workers0.png
                ...

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class:`~tensorbay.dataset.dataset.Dataset` instance.

    """
    dataset = Dataset(DATASET_NAME)
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog.json"))
    segment = dataset.create_segment()
    image_paths = glob(os.path.join(path, "images", "*.png"))
    for image_path in image_paths:
        data = Data(image_path)
        file_name = os.path.splitext(os.path.basename(image_path))[0]
        data.label.box2d = _load_labels(os.path.join(path, "annotations", file_name + ".xml"))
        segment.append(data)
    return dataset


def _load_labels(label_file: str) -> List[LabeledBox2D]:
    label_tree = ElementTree.parse(label_file)
    labels = []
    for obj in label_tree.findall("object"):
        bndbox = obj.find("bndbox")
        labels.append(
            LabeledBox2D(
                *(int(child.text) for child in bndbox),  # type: ignore[arg-type, union-attr]
                category=obj.find("name").text  # type: ignore[union-attr]
            )
        )
    return labels



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""mixin classes for local file and remote file."""

import os
from hashlib import sha1
from http.client import HTTPResponse
from string import printable
from typing import Any, Callable, Dict, Optional
from urllib.parse import quote, urljoin
from urllib.request import pathname2url, urlopen

from _io import BufferedReader

from .repr import ReprMixin


class FileMixin(ReprMixin):
    """FileMixin is a mixin class to mixin file related methods for local file.

    Arguments:
        local_path: The file local path.

    Attributes:
        path: The file local path.

    """

    _checksum: str

    _repr_maxlevel = 3
    _BUFFER_SIZE = 65536

    def __init__(self, local_path: str) -> None:
        self.path = local_path

    def _repr_head(self) -> str:
        return f'{self.__class__.__name__}("{self.path}")'

    def _get_callback_body(self) -> Dict[str, Any]:
        return {"checksum": self.get_checksum(), "fileSize": os.path.getsize(self.path)}

    def get_checksum(self) -> str:
        """Get and cache the sha1 checksum of the local data.

        Returns:
            The sha1 checksum of the local data.

        """
        if not hasattr(self, "_checksum"):
            sha1_object = sha1()
            with open(self.path, "rb") as fp:
                while True:
                    data = fp.read(self._BUFFER_SIZE)
                    if not data:
                        break
                    sha1_object.update(data)

            self._checksum = sha1_object.hexdigest()

        return self._checksum

    def get_url(self) -> str:
        """Return the url of the local data file.

        Returns:
            The url of the local data.

        """
        return urljoin("file:", pathname2url(os.path.abspath(self.path)))

    def open(self) -> BufferedReader:
        """Return the binary file pointer of this file.

        The local file pointer will be obtained by build-in ``open()``.

        Returns:
            The local file pointer for this data.

        """
        return open(self.path, "rb")


class RemoteFileMixin(ReprMixin):
    """RemoteFileMixin is a mixin class to mixin file related methods for remote file.

    Arguments:
        local_path: The file local path.
        _url_getter: The url getter of the remote file.

    Attributes:
        path: The file local path.

    """

    _repr_maxlevel = 3

    def __init__(
        self, remote_path: str, *, _url_getter: Optional[Callable[[str], str]] = None
    ) -> None:
        self.path = remote_path
        self._url_getter = _url_getter

    def _repr_head(self) -> str:
        return f'{self.__class__.__name__}("{self.path}")'

    def get_url(self) -> str:
        """Return the url of the data hosted by tensorbay.

        Returns:
            The url of the data.

        Raises:
            ValueError: When the _url_getter is missing.

        """
        if not self._url_getter:
            raise ValueError(
                f"The file URL cannot be got because {self._repr_head()} has no url getter"
            )

        return self._url_getter(self.path)

    def open(self) -> HTTPResponse:
        """Return the binary file pointer of this file.

        The remote file pointer will be obtained by ``urllib.request.urlopen()``.

        Returns:
            The remote file pointer for this data.

        """
        return urlopen(quote(self.get_url(), safe=printable))  # type: ignore[no-any-return]



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of the Coin Image dataset."""

from .loader import CoinImage

__all__ = ["CoinImage"]



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Implementation of gas ls."""

from typing import Dict, Iterable, Optional

import click

from ..client import GAS
from ..client.dataset import FusionDatasetClient
from ..client.gas import DatasetClientType
from .tbrn import TBRN, TBRNType
from .utility import error, get_dataset_client, get_gas


def _echo_data(
    dataset_name: str,
    draft_number: Optional[int],
    revision: Optional[str],
    segment_name: str,
    data_iter: Iterable[str],
) -> None:
    """Echo files in data_iter under 'tb:dataset_name:segment_name'.

    Arguments:
        dataset_name: The name of the dataset the segment belongs to.
        draft_number: The draft number (if the status is draft).
        revision: The revision (if the status is not draft).
        segment_name: The name of the segment.
        data_iter: Iterable data to be echoed.

    """
    for data in data_iter:
        click.echo(
            TBRN(
                dataset_name,
                segment_name,
                remote_path=data,
                draft_number=draft_number,
                revision=revision,
            ).get_tbrn()
        )


def _validate_dataset_client(dataset_client: DatasetClientType) -> None:
    if not dataset_client.status.draft_number and not dataset_client.status.commit_id:
        error(f'"{dataset_client.name}" has no commit history.')


def _ls_dataset(gas: GAS, info: TBRN, list_all_files: bool, show_total_num: bool) -> None:
    dataset_client = get_dataset_client(gas, info)
    _validate_dataset_client(dataset_client)

    segment_names = dataset_client.list_segment_names()
    if not list_all_files:
        if show_total_num:
            click.echo(f"total {len(segment_names)}")
        for segment_name in segment_names:
            click.echo(TBRN(info.dataset_name, segment_name).get_tbrn())
        return

    if isinstance(dataset_client, FusionDatasetClient):
        error('"-a" flag is not supported for fusion dataset yet')

    segment_paths = [
        dataset_client.get_segment(segment_name).list_data_paths() for segment_name in segment_names
    ]
    if show_total_num:
        total = sum(len(segment_path) for segment_path in segment_paths)
        click.echo(f"total {total}")

    for segment_name, segment_path in zip(segment_names, segment_paths):
        _echo_data(
            info.dataset_name,
            info.draft_number,
            info.revision,
            segment_name,
            segment_path,
        )


def _ls_segment(
    gas: GAS,
    info: TBRN,
    list_all_files: bool,  # pylint: disable=unused-argument
    show_total_num: bool,
) -> None:
    dataset_client = get_dataset_client(gas, info)
    _validate_dataset_client(dataset_client)
    if isinstance(dataset_client, FusionDatasetClient):
        error("List fusion segment is not supported yet")

    segment_path = dataset_client.get_segment(info.segment_name).list_data_paths()
    if show_total_num:
        click.echo(f"total {len(segment_path)}")
    _echo_data(
        info.dataset_name,
        info.draft_number,
        info.revision,
        info.segment_name,
        segment_path,
    )


def _ls_normal_file(
    gas: GAS,  # pylint: disable=unused-argument
    info: TBRN,  # pylint: disable=unused-argument
    list_all_files: bool,  # pylint: disable=unused-argument
    show_total_num: bool,  # pylint: disable=unused-argument
) -> None:
    error("List for specific file is not supported yet")


_LS_FUNCS = {
    TBRNType.DATASET: _ls_dataset,
    TBRNType.SEGMENT: _ls_segment,
    TBRNType.NORMAL_FILE: _ls_normal_file,
}


def _implement_ls(
    obj: Dict[str, str], tbrn: str, list_all_files: bool, show_total_num: bool
) -> None:
    gas = get_gas(**obj)
    if not tbrn:
        dataset_names = gas.list_dataset_names()
        if show_total_num:
            click.echo(f"total {len(dataset_names)}")
        for dataset_name in dataset_names:
            click.echo(TBRN(dataset_name).get_tbrn())
        return

    info = TBRN(tbrn=tbrn)
    _LS_FUNCS[info.type](gas, info, list_all_files, show_total_num)



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Implementation of gas log."""
import bisect
from collections import defaultdict
from datetime import datetime
from itertools import islice, zip_longest
from textwrap import indent
from typing import DefaultDict, Dict, Iterator, List, Optional, Type, Union

import click

from ..client.gas import DatasetClientType
from ..client.struct import Commit
from .auth import INDENT
from .tbrn import TBRN, TBRNType
from .utility import error, get_gas, is_win, shorten

_FULL_LOG = """commit {}
Author: {}
Date: {}

    {}
"""


def _implement_log(  # pylint: disable=too-many-arguments
    obj: Dict[str, str],
    tbrn: str,
    max_count: Optional[int],
    oneline: bool,
    is_all: bool,
    graph: bool,
) -> None:
    gas = get_gas(**obj)
    info = TBRN(tbrn=tbrn)
    if info.type != TBRNType.DATASET:
        error(f'To log commits, "{info}" must be a dataset')

    dataset_client = gas._get_dataset_with_any_type(  # pylint: disable=protected-access
        info.dataset_name
    )
    commit_id_to_branches: DefaultDict[str, List[str]] = defaultdict(list)
    for branch in dataset_client.list_branches():
        commit_id_to_branches[branch.commit_id].append(branch.name)
    if is_all:
        revisions: List[Optional[str]] = [branch.name for branch in dataset_client.list_branches()]
    else:
        revisions = [info.revision] if info.revision else [dataset_client.status.branch_name]

    Printer: Union[Type[_GraphPrinter], Type[_CommitPrinter]] = (
        _GraphPrinter if graph else _CommitPrinter
    )
    commit_generator = islice(
        Printer(dataset_client, revisions, commit_id_to_branches, oneline).generate_commits(),
        max_count,
    )
    if is_win():
        for item in commit_generator:
            click.echo(item)
    else:
        click.echo_via_pager(commit_generator)


def _get_oneline_log(commit: Commit, branch_name: Optional[str]) -> str:
    commit_id = shorten(commit.commit_id)
    if branch_name:
        commit_id = f"{commit_id} ({branch_name})"
    return f"{commit_id} {commit.title}\n"


def _get_full_log(commit: Commit, branch_name: Optional[str]) -> str:
    description = commit.description
    if description:
        description = f"\n\n{indent(description, INDENT)}"
    commit_message = f"{commit.title}{description}\n"
    commit_id = commit.commit_id
    if branch_name:
        commit_id = f"{commit_id} ({branch_name})"
    return _FULL_LOG.format(
        commit_id,
        commit.committer.name,
        datetime.fromtimestamp(commit.committer.date).strftime("%a %b %d %H:%M:%S %y"),
        commit_message,
    )


class _CommitPrinter:
    """This class defines the structure of logging commits.

    Arguments:
        dataset_client: The dataset that needs to be logged.
        revisions: The revisions that needs to be logged.
        oneline: Whether to log with oneline method.
        commit_id_to_branches: The map of commit id to branch name.

    """

    def __init__(
        self,
        dataset_client: DatasetClientType,
        revisions: List[Optional[str]],
        commit_id_to_branches: Dict[str, List[str]],
        oneline: bool,
    ):
        all_commits = list(map(dataset_client.list_commits, revisions))

        if len(all_commits[0]) == 0:
            error(f'Dataset "{dataset_client.name}" has no commit history')

        self._commit_id_to_branches = commit_id_to_branches
        self._printer = _get_oneline_log if oneline else _get_full_log

        # Sort commits from different branches by the date of the latest commit of each branch.
        self._sorted_commits = sorted(all_commits, key=lambda x: x[0].committer.date)
        self._keys = [commits[0].committer.date for commits in self._sorted_commits]

    def _merge(self, latest_commit: Commit) -> bool:
        if len(self._sorted_commits) <= 1:
            return False

        date = latest_commit.committer.date
        commit_id = latest_commit.commit_id
        for commits in islice(reversed(self._sorted_commits), 1, None):
            commit = commits[0]
            # Traverse all the commits with the same timestamp,
            # if the commit id is the same as the latest commit,
            # then merge the branch where the latest commit is located.
            if commit.committer.date != date:
                return False
            # Same commit, merge branches.
            if commit.commit_id == commit_id:
                break
        else:
            return False
        # Merge branches.
        del self._sorted_commits[-1]
        del self._keys[-1]
        return True

    def _sort(self) -> None:
        """Sort commits paging list by commit date."""
        # Only one branch exists.
        if len(self._sorted_commits) == 1:
            return

        # Binary insert.
        commits = self._sorted_commits.pop()
        del self._keys[-1]
        date = commits[0].committer.date
        index = bisect.bisect_left(self._keys, date)
        self._sorted_commits.insert(index, commits)
        self._keys.insert(index, date)

    def _generate_commits(self) -> Iterator[Commit]:
        """Get the latest commit in commit list.

        Yields:
            The latest commit.

        """
        while True:
            try:
                latest_commit = self._sorted_commits[-1].pop(0)
            except IndexError:
                return
            if self._merge(latest_commit):
                continue
            yield latest_commit
            self._sort()

    def generate_commits(self) -> Iterator[str]:
        """Get the the log of all commits.

        Yields:
            The commit log.

        """
        for commit in self._generate_commits():
            commit_id = commit.commit_id
            print_branch_name = ", ".join(self._commit_id_to_branches.get(commit_id, ()))
            yield self._printer(commit, print_branch_name)


class _CommitNode:
    """This class defines the tree struct of graphical logging commits.

    Arguments:
        commit: The commit that needs to be saved in the tree.

    """

    def __init__(self, commit: Commit):
        self.commit = commit
        self.children: List[_CommitNode] = []
        self.available_child_num = 0
        self.parent: Optional[_CommitNode] = None

    def add_child(self, child_node: "_CommitNode") -> None:
        """Save the child node for the parent node.

        Arguments:
            child_node: The child node.

        """
        child_node.parent = self
        self.children.append(child_node)
        self.available_child_num += 1


class _GraphPrinter:
    """This class defines the structure of logging graphical commits stack.

    Arguments:
        dataset_client: The dataset that needs to be logged.
        revisions: The revisions that needs to be logged.
        oneline: Whether to log with oneline method.
        commit_id_to_branches: The map of commit id to branch name.

    """

    def __init__(
        self,
        dataset_client: DatasetClientType,
        revisions: List[Optional[str]],
        commit_id_to_branches: Dict[str, List[str]],
        oneline: bool,
    ):
        self._commit_id_to_branches = commit_id_to_branches
        self._graph_printer = self._add_graph_oneline if oneline else self._add_graph_full

        self._sorted_leaves = self._build_commit_tree(dataset_client, revisions)
        self._layer = 1
        self._pointer = 0
        self._merge_pointer: Optional[int] = None

    def _get_log_node(self) -> _CommitNode:
        """Get the next log commit node.

        Returns:
            The next log commit node.

        Raises:
            RuntimeError: Graphical logging algorithm error.

        """
        self._merge_pointer = None
        for log_node in islice(self._sorted_leaves, self._pointer, None):
            if log_node.available_child_num == 0:
                return log_node

            self._pointer += 1
            if self._pointer >= self._layer:
                self._layer += 1

        raise RuntimeError("Graphical logging algorithm error.")

    def _merge_branches(self, parent: Optional[_CommitNode]) -> int:
        """Merge branches.

        Arguments:
            parent: The parent node.

        Returns:
            The original print pointer.

        """
        if parent not in self._sorted_leaves:
            return self._pointer
        index = self._sorted_leaves.index(parent)
        original_pointer = self._pointer
        self._pointer, self._merge_pointer = sorted((index, self._pointer))
        self._layer -= 1
        del self._sorted_leaves[self._merge_pointer]
        return original_pointer

    def _set_next_node(self, node: _CommitNode) -> None:
        """Set the next node at the position of the printing pointer.

        Arguments:
            node: The next node of the printing pointer.

        """
        node.available_child_num -= 1
        self._sorted_leaves[self._pointer] = node

    def _add_graph_oneline(
        self, commit: Commit, branch_name: Optional[str], original_pointer: int
    ) -> str:
        log = _get_oneline_log(commit, branch_name)
        prefixes = ["|"] * self._layer
        # Don't merge branches.
        if self._merge_pointer is None:
            return f"{self._get_title_prefix(prefixes, original_pointer)} {log}"

        # Merge branches.
        prefixes.append("|")
        lines = [f"{self._get_title_prefix(prefixes, original_pointer)} {log}"]
        lines.extend(f"{prefixes}\n" for prefixes in self._get_merge_prefixes())
        return "".join(lines)

    def _add_graph_full(
        self, commit: Commit, branch_name: Optional[str], original_pointer: int
    ) -> str:
        log = _get_full_log(commit, branch_name)
        splitlines = iter(log.splitlines())
        prefixes = ["|"] * self._layer
        # Don't merge branches.
        if self._merge_pointer is None:
            details_prefix = " ".join(prefixes)
            lines = [f"{self._get_title_prefix(prefixes, original_pointer)} {next(splitlines)}\n"]
            lines.extend(f"{details_prefix} {line}\n" for line in splitlines)
        else:
            # Merge branches.
            prefixes.append("|")
            merge_prefixes = self._get_merge_prefixes()
            lines = [f"{self._get_title_prefix(prefixes, original_pointer)} {next(splitlines)}\n"]
            lines.extend(self._combine_details(list(merge_prefixes), list(splitlines)))

        return "".join(lines)

    def _get_merge_prefixes(self) -> Iterator[str]:
        for i in range(self._merge_pointer, self._pointer, -1):  # type: ignore[arg-type]
            prefixes = ["| "] * self._layer
            prefixes[i - 1] = "|/"
            yield "".join(prefixes)

    @staticmethod
    def _get_title_prefix(prefixes: List[str], original_pointer: int) -> str:
        title_prefixes = prefixes.copy()
        title_prefixes[original_pointer] = "*"
        return " ".join(title_prefixes)

    @staticmethod
    def _combine_details(merge_prefixes: List[str], messages: List[str]) -> Iterator[str]:
        fillvalue = (
            "" if len(merge_prefixes) > len(messages) else merge_prefixes[0].replace("/", " ")
        )
        for prefix, message in zip_longest(merge_prefixes, messages, fillvalue=fillvalue):
            yield f"{prefix}  {message}\n"

    @staticmethod
    def _build_commit_tree(
        dataset_client: DatasetClientType, revisions: List[Optional[str]]
    ) -> List[_CommitNode]:
        commit_to_node: Dict[str, _CommitNode] = {}
        leaves: Dict[str, _CommitNode] = {}
        for revision in revisions:
            commits = dataset_client.list_commits(revision)
            child_node: Optional[_CommitNode] = None
            for index, commit in enumerate(commits):
                commit_id = commit.commit_id
                current_node = commit_to_node.get(commit_id, _CommitNode(commit))
                if child_node:
                    current_node.add_child(child_node)
                # Commit already exists in the tree.
                if commit_id in commit_to_node:
                    break

                # Save leaf node to leaf set.
                if index == 0:
                    leaves[commit_id] = current_node
                # Save commit to commit dict.
                commit_to_node[commit_id] = current_node
                child_node = current_node

        # Check the correction of leaf set.
        for commit_id in tuple(leaves):
            if leaves[commit_id].available_child_num != 0:
                del leaves[commit_id]
        return sorted(leaves.values(), key=lambda x: x.commit.committer.date, reverse=True)

    def generate_commits(self) -> Iterator[str]:
        """Get the graphical commit log.

        Yields:
            The graphical commit log.

        """
        while True:
            current_node = self._get_log_node()
            parent = current_node.parent
            # Merge branch.
            original_pointer = self._merge_branches(parent)

            print_branch_name = ", ".join(
                self._commit_id_to_branches.get(current_node.commit.commit_id, ())
            )
            yield self._graph_printer(current_node.commit, print_branch_name, original_pointer)

            if not parent:
                break
            self._set_next_node(parent)



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Implementation of gas tag."""

from typing import Dict

import click

from ..client.gas import DatasetClientType
from .tbrn import TBRN, TBRNType
from .utility import error, get_dataset_client, get_gas


def _implement_tag(obj: Dict[str, str], tbrn: str, name: str, is_delete: bool) -> None:
    info = TBRN(tbrn=tbrn)

    if info.type != TBRNType.DATASET:
        error(f'To operate a tag, "{info}" must be a dataset')

    gas = get_gas(**obj)
    dataset_client = get_dataset_client(gas, info)

    if is_delete:
        _delete_tag(dataset_client, info)
    elif name:
        _create_tag(dataset_client, name)
    else:
        _list_tags(dataset_client)


def _delete_tag(dataset_client: DatasetClientType, info: TBRN) -> None:
    if not info.revision:
        error(f'To delete a tag, "{info}" must have a tag name')

    dataset_client.delete_tag(info.revision)
    tag_tbrn = TBRN(dataset_client.name, revision=info.revision).get_tbrn()
    click.echo(f"{tag_tbrn} is deleted successfully")


def _create_tag(dataset_client: DatasetClientType, name: str) -> None:
    if dataset_client.status.is_draft:
        error(f'To create a tag, "{dataset_client.name}" cannot be in draft status')

    if not dataset_client.status.commit_id:
        error(f'To create a tag, "{dataset_client.name}" should have commit history')

    dataset_client.create_tag(name=name)
    tag_tbrn = TBRN(dataset_client.name, revision=name).get_tbrn()
    click.echo(f"{tag_tbrn} is created successfully")


def _list_tags(dataset_client: DatasetClientType) -> None:
    for tag in dataset_client.list_tags():
        click.echo(tag.name)



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Geometry related classes."""

from .box import Box2D, Box3D
from .keypoint import Keypoint2D, Keypoints2D
from .polygon import RLE, MultiPolygon, Polygon
from .polyline import MultiPolyline2D, Polyline2D
from .transform import Transform3D
from .vector import Vector, Vector2D, Vector3D

__all__ = [
    "Box2D",
    "Box3D",
    "Keypoint2D",
    "Keypoints2D",
    "Polygon",
    "Polyline2D",
    "MultiPolygon",
    "MultiPolyline2D",
    "RLE",
    "Transform3D",
    "Vector",
    "Vector2D",
    "Vector3D",
]



#!/usr/bin/env python3
#
# Copytright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of the BDD100K dataset."""

from .loader import BDD100K

__all__ = ["BDD100K"]
